# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Fixtures for workflows of trandformation steps."""

from dataclasses import dataclass
from typing import Generic, TypeVar

from pydantic import BaseModel

from metldata.builtin_transformations.delete_slots import SLOT_DELETION_TRANSFORMATION
from metldata.builtin_transformations.infer_references import (
    REFERENCE_INFERENCE_TRANSFORMATION,
)
from metldata.builtin_workflows.ghga_archive import GHGA_ARCHIVE_WORKFLOW
from metldata.custom_types import Json
from metldata.event_handling.models import SubmissionAnnotation
from metldata.model_utils.essentials import MetadataModel
from metldata.transform.base import WorkflowDefinition, WorkflowStep
from tests.fixtures.utils import BASE_DIR, read_yaml

Config = TypeVar("Config", bound=BaseModel)

EXAMPLE_WORKFLOW_DEFINITION = WorkflowDefinition(
    description="A workflow for testing.",
    steps={
        "infer_references": WorkflowStep(
            description="A step for inferring references.",
            transformation_definition=REFERENCE_INFERENCE_TRANSFORMATION,
            input=None,
        ),
        "delete_slots": WorkflowStep(
            description="A step for deleting slots.",
            transformation_definition=SLOT_DELETION_TRANSFORMATION,
            input="infer_references",
        ),
    },
    artifacts={
        "inferred_and_restricted": "infer_references",
        "inferred_and_public": "delete_slots",
    },
)


@dataclass(frozen=True)
class WorkflowTestCase(Generic[Config]):
    """A test case for a workflow."""

    workflow_name: str
    case_name: str
    workflow_definition: WorkflowDefinition
    config: Config
    original_model: MetadataModel
    original_metadata: Json
    submission_annotation: SubmissionAnnotation
    artifact_models: dict[str, MetadataModel]
    artifact_metadata: dict[str, Json]

    def __str__(self) -> str:
        return f"{self.workflow_name}-{self.case_name}"


def _read_test_case(
    *,
    workflow_name: str,
    workflow_definition: WorkflowDefinition,
    case_name: str,
) -> WorkflowTestCase[Config]:
    """Read a test case for a workflow."""

    case_dir = BASE_DIR / "workflows" / workflow_name / case_name
    config_path = case_dir / "config.yaml"
    original_model_path = case_dir / "original_model.yaml"
    original_metadata_path = case_dir / "original_metadata.yaml"
    metadata_annotation_path = case_dir / "metadata_annotation.yaml"
    artifacts_dir = case_dir / "artifacts"

    artifact_models: dict[str, MetadataModel] = {}
    artifact_metadata: dict[str, Json] = {}

    for artifact in workflow_definition.artifacts:
        artifact_dir = artifacts_dir / artifact
        if artifact_dir.exists():
            model_path = artifact_dir / "transformed_model.yaml"
            if model_path.exists():
                artifact_models[artifact] = MetadataModel.init_from_path(model_path)
            metadata_path = artifact_dir / "transformed_metadata.yaml"
            if metadata_path.exists():
                artifact_metadata[artifact] = read_yaml(metadata_path)

    return WorkflowTestCase(
        workflow_name=workflow_name,
        case_name=case_name,
        workflow_definition=workflow_definition,
        config=workflow_definition.config_cls(**read_yaml(config_path)),  # type: ignore
        original_model=MetadataModel.init_from_path(original_model_path),
        original_metadata=read_yaml(original_metadata_path),
        submission_annotation=(
            SubmissionAnnotation(**read_yaml(metadata_annotation_path))
            if metadata_annotation_path.exists()
            else SubmissionAnnotation(accession_map={})
        ),
        artifact_models=artifact_models,
        artifact_metadata=artifact_metadata,
    )


def _read_all_test_cases_for_a_workflow(
    *,
    workflow_name: str,
    workflow_definition: WorkflowDefinition,
) -> list[WorkflowTestCase]:
    """Read all test cases for a workflow."""

    base_dir = BASE_DIR / "workflows" / workflow_name
    case_names = [path.name for path in base_dir.iterdir() if path.is_dir()]

    return [
        _read_test_case(
            workflow_name=workflow_name,
            workflow_definition=workflow_definition,
            case_name=case_name,
        )
        for case_name in case_names
    ]


def _read_all_test_cases(
    *, workflows_by_name: dict[str, WorkflowDefinition]
) -> list[WorkflowTestCase]:
    """Read all test cases for the specified workflows."""

    return [
        test_case
        for workflow_name, workflow_definition in workflows_by_name.items()
        for test_case in _read_all_test_cases_for_a_workflow(
            workflow_name=workflow_name,
            workflow_definition=workflow_definition,
        )
    ]


WORKFLOWS_BY_NAME: dict[str, WorkflowDefinition] = {
    "example_workflow": EXAMPLE_WORKFLOW_DEFINITION,
    "ghga_archive_workflow": GHGA_ARCHIVE_WORKFLOW,
}

WORKFLOW_TEST_CASES = _read_all_test_cases(workflows_by_name=WORKFLOWS_BY_NAME)

EXAMPLE_WORKFLOW_TEST_CASE = [
    test_case
    for test_case in WORKFLOW_TEST_CASES
    if test_case.workflow_name == "example_workflow"
][0]
EXAMPLE_ARTIFACT_MODELS = EXAMPLE_WORKFLOW_TEST_CASE.artifact_models
EXAMPLE_ARTIFACTS = EXAMPLE_WORKFLOW_TEST_CASE.artifact_metadata



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for transforming metadata."""


from metldata.metadata_utils import get_resources_of_class, upsert_resources_in_metadata
from metldata.model_utils.anchors import AnchorPoint
from metldata.transform.base import Json, MetadataModelTransformationError


def delete_slot_from_resource(resource: Json, slot_name: str) -> Json:
    """Delete a slot from a resource. Returns a modified copy of the resource."""

    if slot_name not in resource:
        raise MetadataModelTransformationError(
            f"Slot '{slot_name}' not found in resource '{resource}'"
        )

    modified_resource = resource.copy()
    del modified_resource[slot_name]
    return modified_resource


def delete_class_slot(
    metadata: Json,
    class_name: str,
    slot_name: str,
    anchor_points_by_target: dict[str, AnchorPoint],
) -> Json:
    """Delete a slot from a class in the provided metadata. Returns a modified copy of
    the metadata."""

    resources = get_resources_of_class(
        class_name=class_name,
        global_metadata=metadata,
        anchor_points_by_target=anchor_points_by_target,
    )

    modified_resources = [
        delete_slot_from_resource(resource=resource, slot_name=slot_name)
        for resource in resources
    ]

    return upsert_resources_in_metadata(
        resources=modified_resources,
        class_name=class_name,
        global_metadata=metadata,
        anchor_points_by_target=anchor_points_by_target,
    )


def delete_class_slots(
    metadata: Json,
    slots_to_delete: dict[str, list[str]],
    anchor_points_by_target: dict[str, AnchorPoint],
) -> Json:
    """Delete slots from provided metadata. Returns a modified copy of the metadata."""

    for class_name, slot_names in slots_to_delete.items():
        for slot_name in slot_names:
            metadata = delete_class_slot(
                metadata=metadata,
                class_name=class_name,
                slot_name=slot_name,
                anchor_points_by_target=anchor_points_by_target,
            )

    return metadata



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Example of ArtifactInfo objects."""

from metldata.artifacts_rest.artifact_info import load_artifact_info
from tests.fixtures.metadata_models import VALID_MINIMAL_METADATA_MODEL
from tests.fixtures.workflows import EXAMPLE_ARTIFACT_MODELS

# artifact info for the minimal valid metadata model:
MINIMAL_ARTIFACT_INFO = load_artifact_info(
    name="valid_minimal_metadata_model",
    description="This is a minimal valid metadata model.",
    model=VALID_MINIMAL_METADATA_MODEL,
)

# artifact infos for the example workflow:
EXAMPLE_ARTIFACT_INFOS = [
    load_artifact_info(
        name=artifact_name, description=artifact_name, model=artifact_model
    )
    for artifact_name, artifact_model in EXAMPLE_ARTIFACT_MODELS.items()
]



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test model functionality."""

from datetime import timedelta

from ghga_service_commons.utils.utc_dates import now_as_utc

from metldata.submission_registry.models import (
    StatusChange,
    Submission,
    SubmissionStatus,
)


def test_submission_current_status():
    """Test whether the current status property returns the latest status."""

    submission = Submission(
        title="test",
        description="test",
        content={"test_class": [{"alias": "test_alias1"}]},
        accession_map={"test_class": {"test_alias1": "test_accession1"}},
        id="testsubmission001",
        status_history=(
            StatusChange(
                timestamp=now_as_utc(), new_status=SubmissionStatus.IN_REVIEW  # second
            ),
            StatusChange(
                timestamp=now_as_utc() + timedelta(days=10),  # third
                new_status=SubmissionStatus.COMPLETED,
            ),
            StatusChange(
                timestamp=now_as_utc() - timedelta(days=10),  # first
                new_status=SubmissionStatus.PENDING,
            ),
        ),
    )

    assert submission.current_status == SubmissionStatus.COMPLETED



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test the client module."""

import json
from uuid import uuid4

import pytest
from pytest_httpx import HTTPXMock

from metldata.load.client import upload_artifacts_via_http_api
from metldata.load.collect import get_artifact_topic
from metldata.load.config import ArtifactLoaderClientConfig
from metldata.load.models import ArtifactResourceDict
from tests.fixtures.event_handling import file_system_event_fixture  # noqa: F401
from tests.fixtures.event_handling import Event, FileSystemEventFixture

EXAMPLE_ARTIFACTS: ArtifactResourceDict = {
    "example_artifact": [
        {"some": "example", "artifact": "data"},
    ]
}


@pytest.mark.asyncio
async def test_upload_artifacts_via_http_api(
    file_system_event_fixture: FileSystemEventFixture,  # noqa: F811
    httpx_mock: HTTPXMock,
):
    """Test the happy path of using the upload_artifacts_via_http_api function."""

    token = "some-token"
    config = ArtifactLoaderClientConfig(
        artifact_types=list(EXAMPLE_ARTIFACTS.keys()),
        artifact_topic_prefix="artifact",
        loader_api_root="http://localhost:8000",
        **file_system_event_fixture.config.dict(),
    )

    # publish artifacts:
    artifact_events = [
        Event(
            topic=get_artifact_topic(
                artifact_topic_prefix=config.artifact_topic_prefix,
                artifact_type=artifact_type,
            ),
            type_=artifact_type,
            key=str(uuid4()),  # will later be the submission id
            payload=artifact_content,
        )
        for artifact_type, artifact_contents in EXAMPLE_ARTIFACTS.items()
        for artifact_content in artifact_contents
    ]
    await file_system_event_fixture.publish_events(artifact_events)

    # mock the api:
    httpx_mock.add_response(
        url=f"{config.loader_api_root}/rpc/load-artifacts",
        method="POST",
        status_code=204,
    )

    # upload to api:
    upload_artifacts_via_http_api(token=token, config=config)

    # ensure that the api was called with the expected data:
    observed_requests = httpx_mock.get_requests()
    assert len(observed_requests) == 1
    observed_artifacts = json.loads(observed_requests[0].content.decode("utf-8"))
    assert observed_artifacts == EXAMPLE_ARTIFACTS
    assert observed_requests[0].headers["Authorization"] == f"Bearer {token}"



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for loading artifact resources."""


from metldata.artifacts_rest.artifact_dao import ArtifactDaoCollection
from metldata.artifacts_rest.load_resources import load_artifact_resources
from metldata.artifacts_rest.models import ArtifactInfo
from metldata.load.models import ArtifactResourceDict


class ArtifactResourcesInvalid(RuntimeError):
    """Raised when artifact resources are invalid."""


def check_artifact_resources(
    artifact_resources: ArtifactResourceDict, artifact_infos: dict[str, ArtifactInfo]
) -> None:
    """Check the provided artifact resources against the given artifact info.

    Raises:
        ArtifactResourcesInvalid: If the artifact resources are invalid.
    """

    for artifact_name in artifact_resources:
        if artifact_name not in artifact_infos:
            raise ArtifactResourcesInvalid(f"Artifact '{artifact_name}' is unknown.")


async def load_artifacts_using_dao(
    artifact_resources: ArtifactResourceDict,
    artifact_info_dict: dict[str, ArtifactInfo],
    dao_collection: ArtifactDaoCollection,
) -> None:
    """Load artifact resources from multiple submissions using the given dao collection."""

    for artifact_name, artifact_contents in artifact_resources.items():
        for artifact_content in artifact_contents:
            await load_artifact_resources(
                artifact_content=artifact_content,
                artifact_info=artifact_info_dict[artifact_name],
                dao_collection=dao_collection,
            )



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test the submission registry."""

import pytest

from metldata.accession_registry.accession_registry import AccessionRegistry
from metldata.accession_registry.accession_store import AccessionStore
from metldata.config import SubmissionConfig
from metldata.event_handling.event_handling import FileSystemEventPublisher
from metldata.submission_registry import models
from metldata.submission_registry.event_publisher import SourceEventPublisher
from metldata.submission_registry.submission_registry import SubmissionRegistry
from metldata.submission_registry.submission_store import SubmissionStore
from tests.fixtures.config import config_sub_fixture  # noqa: F401
from tests.fixtures.event_handling import file_system_event_fixture  # noqa: F401
from tests.fixtures.event_handling import FileSystemEventFixture
from tests.fixtures.metadata import (
    INVALID_MINIMAL_METADATA_EXAMPLES,
    VALID_MINIMAL_METADATA_EXAMPLES,
)
from tests.submission_registry.test_event_publisher import check_source_events


def test_happy(
    config_sub_fixture: SubmissionConfig,  # noqa: F811
    file_system_event_fixture: FileSystemEventFixture,  # noqa: F811
):
    """Test the happy path of using the submission registry."""

    # inject dependencies:
    submission_store = SubmissionStore(config=config_sub_fixture)
    provider = FileSystemEventPublisher(config=file_system_event_fixture.config)
    event_publisher = SourceEventPublisher(config=config_sub_fixture, provider=provider)
    accession_store = AccessionStore(config=config_sub_fixture)
    accession_registry = AccessionRegistry(
        config=config_sub_fixture, accession_store=accession_store
    )
    submission_registry = SubmissionRegistry(
        config=config_sub_fixture,
        submission_store=submission_store,
        event_publisher=event_publisher,
        accession_registry=accession_registry,
    )

    # create a new submission:
    submission_header = models.SubmissionHeader(title="test", description="test")
    submission_id = submission_registry.init_submission(header=submission_header)

    # check submission in store:
    observed_submission = submission_store.get_by_id(submission_id)
    assert observed_submission.title == submission_header.title
    assert observed_submission.description == submission_header.description
    assert observed_submission.content is None
    assert observed_submission.current_status == models.SubmissionStatus.PENDING

    # provide content:
    submission_content = VALID_MINIMAL_METADATA_EXAMPLES[0]
    submission_registry.upsert_submission_content(
        submission_id=submission_id, content=submission_content
    )

    # check submission in store:
    observed_submission = submission_store.get_by_id(submission_id)
    assert observed_submission.content == submission_content
    assert observed_submission.current_status == models.SubmissionStatus.PENDING

    # update content:
    submission_content_updated = VALID_MINIMAL_METADATA_EXAMPLES[1]
    submission_registry.upsert_submission_content(
        submission_id=submission_id, content=submission_content_updated
    )

    # check submission in store:
    observed_submission = submission_store.get_by_id(submission_id)
    assert observed_submission.content == submission_content_updated
    assert observed_submission.current_status == models.SubmissionStatus.PENDING

    # check accessions are in store:
    for class_ in observed_submission.accession_map.values():
        for accession in class_.values():
            assert accession_store.exists(accession=accession)

    # check published source event:
    check_source_events(
        expected_submissions=[observed_submission],
        source_event_topic=config_sub_fixture.source_event_topic,
        source_event_type=config_sub_fixture.source_event_type,
        file_system_event_fixture=file_system_event_fixture,
    )

    # declare submission is complete:
    submission_registry.complete_submission(id_=submission_id)

    # check submission in store:
    observed_submission = submission_store.get_by_id(submission_id)
    assert observed_submission.current_status == models.SubmissionStatus.COMPLETED

    # check published source event:
    check_source_events(
        expected_submissions=[observed_submission],
        source_event_topic=config_sub_fixture.source_event_topic,
        source_event_type=config_sub_fixture.source_event_type,
        file_system_event_fixture=file_system_event_fixture,
    )


def test_failed_content_validation(
    config_sub_fixture: SubmissionConfig,  # noqa: F811
    file_system_event_fixture: FileSystemEventFixture,  # noqa: F811
):
    """Test that invalid content cannot be upserted."""

    # inject dependencies:
    submission_store = SubmissionStore(config=config_sub_fixture)
    provider = FileSystemEventPublisher(config=file_system_event_fixture.config)
    event_publisher = SourceEventPublisher(config=config_sub_fixture, provider=provider)
    accession_store = AccessionStore(config=config_sub_fixture)
    accession_registry = AccessionRegistry(
        config=config_sub_fixture, accession_store=accession_store
    )
    submission_registry = SubmissionRegistry(
        config=config_sub_fixture,
        submission_store=submission_store,
        event_publisher=event_publisher,
        accession_registry=accession_registry,
    )

    # create a new submission:
    submission_header = models.SubmissionHeader(title="test", description="test")
    submission_id = submission_registry.init_submission(header=submission_header)

    # check submission in store:
    observed_submission_original = submission_store.get_by_id(submission_id)

    # provide invalid content:
    submission_content = INVALID_MINIMAL_METADATA_EXAMPLES[0]
    with pytest.raises(SubmissionRegistry.ValidationError):
        submission_registry.upsert_submission_content(
            submission_id=submission_id, content=submission_content
        )

    # check that the submission was not changed:
    observed_submission_updated = submission_store.get_by_id(submission_id)
    assert observed_submission_original == observed_submission_updated


def test_update_after_completion(
    config_sub_fixture: SubmissionConfig,  # noqa: F811
    file_system_event_fixture: FileSystemEventFixture,  # noqa: F811
):
    """Test no updates can be carried out after completion."""

    # inject dependencies:
    submission_store = SubmissionStore(config=config_sub_fixture)
    provider = FileSystemEventPublisher(config=file_system_event_fixture.config)
    event_publisher = SourceEventPublisher(config=config_sub_fixture, provider=provider)
    accession_store = AccessionStore(config=config_sub_fixture)
    accession_registry = AccessionRegistry(
        config=config_sub_fixture, accession_store=accession_store
    )
    submission_registry = SubmissionRegistry(
        config=config_sub_fixture,
        submission_store=submission_store,
        event_publisher=event_publisher,
        accession_registry=accession_registry,
    )

    # create a new submission:
    submission_header = models.SubmissionHeader(title="test", description="test")
    submission_id = submission_registry.init_submission(header=submission_header)

    # provide content:
    submission_content = VALID_MINIMAL_METADATA_EXAMPLES[0]
    submission_registry.upsert_submission_content(
        submission_id=submission_id, content=submission_content
    )

    # check submission in store:
    observed_submission = submission_store.get_by_id(submission_id)

    # declare submission is complete:
    submission_registry.complete_submission(id_=submission_id)

    # try to update content:
    submission_content_updated = VALID_MINIMAL_METADATA_EXAMPLES[1]
    with pytest.raises(SubmissionRegistry.StatusError):
        submission_registry.upsert_submission_content(
            submission_id=submission_id, content=submission_content_updated
        )

    # check that the submission content did not change:
    observed_submission = submission_store.get_by_id(submission_id)
    assert observed_submission.content == submission_content



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test the essentials module."""

from copy import deepcopy

from linkml_runtime import SchemaView

from metldata.model_utils.essentials import MetadataModel
from tests.fixtures.metadata_models import (
    VALID_MINIMAL_METADATA_MODEL,
    VALID_MINIMAL_MODEL_EXAMPLE_PATH,
)


def test_metadata_model_from_path():
    """Test the MetadataModel creation from path."""

    model = MetadataModel.init_from_path(VALID_MINIMAL_MODEL_EXAMPLE_PATH)

    assert model.name == "Minimal-Model"


def test_metadata_model_get_schema_view():
    """Test getting a schema view from a MetadataModel."""

    schema_view = VALID_MINIMAL_METADATA_MODEL.schema_view

    assert isinstance(schema_view, SchemaView)


def test_metadata_model_temporary_yaml():
    """Test getting a temporary yaml file from a MetadataModel."""

    model = deepcopy(VALID_MINIMAL_METADATA_MODEL)

    # modify the model:
    model.name = "Test-Model-Modified"

    # get a temporary yaml file:
    with model.temporary_yaml_path() as model_path:
        # load a copy of that file into a new model instance:
        model_copy = MetadataModel.init_from_path(model_path)

    assert model_copy == model



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Config parameters and their defaults."""

from ghga_service_commons.api import ApiConfigBase
from hexkit.providers.mongodb import MongoDbConfig
from pydantic import Field

from metldata.artifacts_rest.config import ArtifactsRestConfig
from metldata.event_handling.event_handling import FileSystemEventConfig
from metldata.load.collect import ArtifactCollectorConfig


class ArtifactLoaderAPIConfig(ArtifactsRestConfig, ApiConfigBase, MongoDbConfig):
    """Config settings for the loader API."""

    loader_token_hashes: list[str] = Field(
        ..., description="Hashes of tokens used to authenticate for loading artifact."
    )


class ArtifactLoaderClientConfig(ArtifactCollectorConfig, FileSystemEventConfig):
    """Config settings for the loader client."""

    loader_api_root: str = Field(..., description="Root URL of the loader API.")



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test the base module."""


import pytest
from pydantic import ValidationError

from metldata.builtin_transformations.delete_slots import SLOT_DELETION_TRANSFORMATION
from metldata.builtin_transformations.infer_references import (
    REFERENCE_INFERENCE_TRANSFORMATION,
)
from metldata.transform.base import WorkflowDefinition, WorkflowStep
from tests.fixtures.workflows import EXAMPLE_WORKFLOW_DEFINITION


def test_workflow_definition_invalid_step_refs():
    """Test that an invalid step reference raises an error."""
    with pytest.raises(ValidationError):
        WorkflowDefinition(
            description="A workflow for testing.",
            steps={
                "infer_references": WorkflowStep(
                    description="A step for inferring references.",
                    transformation_definition=REFERENCE_INFERENCE_TRANSFORMATION,
                    input=None,
                ),
                "delete_slots": WorkflowStep(
                    description="A step for deleting slots.",
                    transformation_definition=SLOT_DELETION_TRANSFORMATION,
                    input="non_existing_step",
                ),
            },
            artifacts={
                "inferred_and_restricted": "infer_references",
                "inferred_and_public": "delete_slots",
            },
        )


def test_workflow_definition_invalid_multiple_first_steps():
    """Test that specifing multiple steps without input raises an exeception."""
    with pytest.raises(ValidationError):
        WorkflowDefinition(
            description="A workflow for testing.",
            steps={
                "infer_references": WorkflowStep(
                    description="A step for inferring references.",
                    transformation_definition=REFERENCE_INFERENCE_TRANSFORMATION,
                    input=None,
                ),
                "delete_slots": WorkflowStep(
                    description="A step for deleting slots.",
                    transformation_definition=SLOT_DELETION_TRANSFORMATION,
                    input=None,
                ),
            },
            artifacts={
                "inferred_and_restricted": "infer_references",
                "inferred_and_public": "delete_slots",
            },
        )


def test_workflow_definition_invalid_artifacts():
    """Test that artifacts referencing non-existing steps raise an exception."""
    with pytest.raises(ValidationError):
        WorkflowDefinition(
            description="A workflow for testing.",
            steps={
                "infer_references": WorkflowStep(
                    description="A step for inferring references.",
                    transformation_definition=REFERENCE_INFERENCE_TRANSFORMATION,
                    input=None,
                ),
                "delete_slots": WorkflowStep(
                    description="A step for deleting slots.",
                    transformation_definition=SLOT_DELETION_TRANSFORMATION,
                    input=None,
                ),
            },
            artifacts={
                "inferred_and_restricted": "non_existing_step",
                "inferred_and_public": "delete_slots",
            },
        )


def test_workflow_definition_config_cls():
    """Test that the config_cls of the WorkflowDefinition generates a concatenated
    config class correctly."""

    config_fields = EXAMPLE_WORKFLOW_DEFINITION.config_cls.__fields__

    assert "infer_references" in config_fields
    assert "delete_slots" in config_fields
    assert (
        config_fields["infer_references"].type_
        == REFERENCE_INFERENCE_TRANSFORMATION.config_cls
    )
    assert (
        config_fields["delete_slots"].type_ == SLOT_DELETION_TRANSFORMATION.config_cls
    )


def test_workflow_definition_step_order_happy():
    """Test that the step order is correctly inferred from the workflow definition."""

    workflow_definition = WorkflowDefinition(
        description="A workflow for testing.",
        steps={
            "step3": WorkflowStep(
                description="A test step.",
                transformation_definition=SLOT_DELETION_TRANSFORMATION,
                input="step2",
            ),
            "step2": WorkflowStep(
                description="A test step.",
                transformation_definition=SLOT_DELETION_TRANSFORMATION,
                input="step1",
            ),
            "step1": WorkflowStep(
                description="A test step.",
                transformation_definition=SLOT_DELETION_TRANSFORMATION,
                input=None,
            ),
            "step4": WorkflowStep(
                description="A test step.",
                transformation_definition=SLOT_DELETION_TRANSFORMATION,
                input="step2",
            ),
        },
        artifacts={
            "output3": "step3",
            "output4": "step4",
        },
    )

    assert workflow_definition.step_order in (
        [
            "step1",
            "step2",
            "step3",
            "step4",
        ],
        [
            "step1",
            "step2",
            "step4",
            "step3",
        ],
    )


def test_workflow_definition_step_order_circular():
    """Test that the step order is correctly inferred from the workflow definition."""

    workflow_definition = WorkflowDefinition(
        description="A workflow for testing.",
        steps={
            "step1": WorkflowStep(
                description="A test step.",
                transformation_definition=SLOT_DELETION_TRANSFORMATION,
                input=None,
            ),
            "step2": WorkflowStep(
                description="A test step.",
                transformation_definition=SLOT_DELETION_TRANSFORMATION,
                input="step4",
            ),
            "step3": WorkflowStep(
                description="A test step.",
                transformation_definition=SLOT_DELETION_TRANSFORMATION,
                input="step2",
            ),
            "step4": WorkflowStep(
                description="A test step.",
                transformation_definition=SLOT_DELETION_TRANSFORMATION,
                input="step3",
            ),
        },
        artifacts={
            "output3": "step3",
            "output4": "step4",
        },
    )

    with pytest.raises(RuntimeError):
        workflow_definition.step_order



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Testing the metadata validator."""

from contextlib import nullcontext

import pytest

from metldata.model_utils.assumptions import MetadataModelAssumptionError
from metldata.model_utils.config import MetadataModelConfig
from metldata.model_utils.essentials import MetadataModel
from tests.fixtures.metadata_models import (
    INVALID_METADATA_MODELS,
    VALID_METADATA_MODELS,
)


@pytest.mark.parametrize(
    "model, is_valid",
    [(valid_model, True) for valid_model in VALID_METADATA_MODELS]
    + [(invalid_model, False) for invalid_model in INVALID_METADATA_MODELS],
)
def test_config(model: MetadataModel, is_valid: bool):
    """Test with valid and invalid model."""

    with model.temporary_yaml_path() as model_path:
        with nullcontext() if is_valid else pytest.raises(  # type:ignore
            MetadataModelAssumptionError
        ):
            config = MetadataModelConfig(metadata_model_path=model_path)
            _ = config.metadata_model



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Testing the event publisher."""

import json

from ghga_service_commons.utils.utc_dates import now_as_utc

from metldata.config import SubmissionConfig
from metldata.event_handling.event_handling import FileSystemEventPublisher
from metldata.event_handling.models import SubmissionAnnotation, SubmissionEventPayload
from metldata.submission_registry import models
from metldata.submission_registry.event_publisher import SourceEventPublisher
from tests.fixtures.config import config_sub_fixture  # noqa: F401
from tests.fixtures.event_handling import file_system_event_fixture  # noqa: F401
from tests.fixtures.event_handling import Event, FileSystemEventFixture


def check_source_events(
    *,
    expected_submissions: list[models.Submission],
    source_event_topic: str,
    source_event_type: str,
    file_system_event_fixture: FileSystemEventFixture,  # noqa: F811
):
    """Check the content of a source event.

    Raises:
        AssertionError: if it does not match the expectations.
        EventNotPublishedError: if the event was not yet published.
    """

    expected_events = [
        Event(
            topic=source_event_topic,
            type_=source_event_type,
            key=expected_submission.id,
            payload=json.loads(
                SubmissionEventPayload(
                    submission_id=expected_submission.id,
                    content=expected_submission.content,
                    annotation=SubmissionAnnotation(
                        accession_map=expected_submission.accession_map
                    ),
                ).json()
            ),
        )
        for expected_submission in expected_submissions
    ]

    file_system_event_fixture.expect_events(expected_events=expected_events)


def test_happy(
    config_sub_fixture: SubmissionConfig,  # noqa: F811
    file_system_event_fixture: FileSystemEventFixture,  # noqa: F811
):
    """Test the happy path of publishing a submission."""

    provider = FileSystemEventPublisher(config=file_system_event_fixture.config)
    event_publisher = SourceEventPublisher(config=config_sub_fixture, provider=provider)

    # publish a submission:
    submission = models.Submission(
        id="submission001",
        title="test",
        description="test",
        content={"test_class": [{"alias": "test_alias1"}]},
        accession_map={"test_class": {"test_alias1": "test_accession1"}},
        status_history=(
            models.StatusChange(
                timestamp=now_as_utc(), new_status=models.SubmissionStatus.PENDING
            ),
        ),
    )
    event_publisher.publish_submission(submission)

    # check published source event:
    check_source_events(
        expected_submissions=[submission],
        source_event_topic=config_sub_fixture.source_event_topic,
        source_event_type=config_sub_fixture.source_event_type,
        file_system_event_fixture=file_system_event_fixture,
    )



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Main logic for running a transformation workflow on submissions."""


from typing import Awaitable, Callable

from metldata.event_handling.event_handling import (
    FileSystemEventPublisher,
    FileSystemEventSubscriber,
)
from metldata.event_handling.models import SubmissionEventPayload
from metldata.model_utils.essentials import MetadataModel
from metldata.transform.artifact_publisher import ArtifactEvent, ArtifactEventPublisher
from metldata.transform.base import WorkflowConfig, WorkflowDefinition
from metldata.transform.config import TransformationEventHandlingConfig
from metldata.transform.handling import WorkflowHandler
from metldata.transform.source_event_subscriber import SourceEventSubscriber


async def run_workflow_on_source_event(
    source_event: SubmissionEventPayload,
    workflow_handler: WorkflowHandler,
    publish_artifact_func: Callable[[ArtifactEvent], Awaitable[None]],
) -> None:
    """Run a transformation workflow on a source event and publish the artifact using
    the provided artifact publisher.

    Args:
        source_event:
            The source event corresponding to one submission and its cotent.
        workflow_handler:
            The workflow handler preconfigured with a workflow definition, a workflow
            config, and the original model of the source events.
        publish_artifact_func: A function for publishing artifacts.
    """

    artifacts = workflow_handler.run(
        metadata=source_event.content, annotation=source_event.annotation
    )

    for artifact_type, artifact_content in artifacts.items():
        artifact_event = ArtifactEvent(
            artifact_type=artifact_type,
            payload=source_event.copy(update={"content": artifact_content}),
        )

        await publish_artifact_func(artifact_event)


async def run_workflow_on_all_source_events(
    *,
    event_config: TransformationEventHandlingConfig,
    workflow_definition: WorkflowDefinition,
    worflow_config: WorkflowConfig,
    original_model: MetadataModel
):
    """Run a subscriber to hand source events to a transformation workflow and
    run a publisher for publishing artifacts."""

    workflow_handler = WorkflowHandler(
        workflow_definition=workflow_definition,
        workflow_config=worflow_config,
        original_model=original_model,
    )
    event_publisher = FileSystemEventPublisher(config=event_config)
    artifact_publisher = ArtifactEventPublisher(
        config=event_config, provider=event_publisher
    )
    source_event_subscriber = SourceEventSubscriber(
        config=event_config,
        run_workflow_func=lambda source_event: run_workflow_on_source_event(
            workflow_handler=workflow_handler,
            source_event=source_event,
            publish_artifact_func=artifact_publisher.publish_artifact,
        ),
    )
    event_subscriber = FileSystemEventSubscriber(
        config=event_config, translator=source_event_subscriber
    )
    await event_subscriber.run()



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for transforming metadata models."""

from linkml_runtime.linkml_model.meta import AnonymousSlotExpression, SlotDefinition

from metldata.builtin_transformations.merge_slots.models import SlotMergeInstruction
from metldata.model_utils.essentials import ExportableSchemaView, MetadataModel
from metldata.model_utils.manipulate import upsert_class_slot
from metldata.transform.base import MetadataModelTransformationError


def get_source_range(
    *, schema_view: ExportableSchemaView, class_name: str, source_slot: str
) -> str:
    """Get the range of a source slot for the provided class.

    Raises:
        MetadataModelTransformationError:
            If the slot has no defined range, or if the slot is using a union range
            with the `all_of` or `any_of` properties.
    """

    slot = schema_view.induced_slot(class_name=class_name, slot_name=source_slot)

    if not slot.range:
        raise MetadataModelTransformationError(
            f"Source slot {source_slot} of class {class_name} has no defined range."
        )

    if slot.all_of or slot.any_of:
        raise MetadataModelTransformationError(
            f"Source slot {source_slot} of class {class_name} is using a union range."
        )

    return str(slot.range)


def get_source_ranges(
    *, schema_view: ExportableSchemaView, class_name: str, source_slots: list[str]
) -> list[str]:
    """Get the ranges of the given source slots."""

    source_ranges: list[str] = []
    for source_slot in source_slots:
        source_range = get_source_range(
            schema_view=schema_view, class_name=class_name, source_slot=source_slot
        )
        if source_range not in source_ranges:
            source_ranges.append(source_range)

    return source_ranges


def is_source_slot_inlined(
    *, schema_view: ExportableSchemaView, class_name: str, slot_name: str
) -> tuple[bool, bool]:
    """Check if a class slot is inlined. Returns a tuple of booleans where the first
    element indicates if the slot is inlined, and the second element indicates if the
    slot is inlined_as_list"""

    slot = schema_view.induced_slot(class_name=class_name, slot_name=slot_name)

    return bool(slot.inlined), bool(slot.inlined_as_list)


def are_source_slots_inlined(
    *, schema_view: ExportableSchemaView, class_name: str, source_slots: list[str]
) -> tuple[bool, bool]:
    """Check if all source slots are inlined. Returns a tuple of booleans where the
    first element indicates if the slots are inlined, and the second element indicates
    if the slots are inlined_as_list"""

    # inspect first slot:
    inlined, inlined_as_list = is_source_slot_inlined(
        schema_view=schema_view, class_name=class_name, slot_name=source_slots[0]
    )

    # make sure that all other slots are identical:
    for source_slot in source_slots[1:]:
        other_inlined, other_inlined_as_list = is_source_slot_inlined(
            schema_view=schema_view, class_name=class_name, slot_name=source_slot
        )

        if inlined != other_inlined or inlined_as_list != other_inlined_as_list:
            raise MetadataModelTransformationError(
                f"Source slots {source_slots} of class {class_name} are not identical."
                + " They differ in their inlined and/or inlined_as_list properties."
            )

    return inlined, inlined_as_list


def get_target_slot(
    *, schema_view: ExportableSchemaView, merge_instruction: SlotMergeInstruction
) -> SlotDefinition:
    """Get a definition of the target slot."""

    source_ranges = get_source_ranges(
        schema_view=schema_view,
        class_name=merge_instruction.class_name,
        source_slots=merge_instruction.source_slots,
    )

    inlined, inlined_as_list = are_source_slots_inlined(
        schema_view=schema_view,
        class_name=merge_instruction.class_name,
        source_slots=merge_instruction.source_slots,
    )

    target_slot_definition = SlotDefinition(
        name=merge_instruction.target_slot,
        required=True,
        multivalued=True,
        description=merge_instruction.target_description,
    )

    if len(source_ranges) == 1:
        target_slot_definition.range = source_ranges.pop()
    else:
        target_slot_definition.any_of = [
            AnonymousSlotExpression(range=source_range)
            for source_range in source_ranges
        ]

    target_slot_definition.inlined = inlined
    if inlined_as_list:
        target_slot_definition.inlined_as_list = inlined_as_list

    return target_slot_definition


def apply_merge_instruction(
    *, schema_view: ExportableSchemaView, merge_instruction: SlotMergeInstruction
) -> ExportableSchemaView:
    """Apply the provided merge instructions to the provided schema_view."""

    target_slot_definition = get_target_slot(
        schema_view=schema_view, merge_instruction=merge_instruction
    )

    return upsert_class_slot(
        schema_view=schema_view,
        class_name=merge_instruction.class_name,
        new_slot=target_slot_definition,
    )


def merge_slots_in_model(
    *, model: MetadataModel, merge_instructions: list[SlotMergeInstruction]
) -> MetadataModel:
    """Merge slots in the provided model according to the provided instructions."""

    schema_view = model.schema_view

    for merge_instruction in merge_instructions:
        schema_view = apply_merge_instruction(
            schema_view=schema_view, merge_instruction=merge_instruction
        )

    return schema_view.export_model()



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Reference models."""

from pydantic import BaseModel, Field, root_validator

from metldata.builtin_transformations.infer_references.path.path import ReferencePath


class ReferenceDetails(BaseModel):
    """A base model for describing an inferred reference that is based on existing
    references."""

    path: ReferencePath = Field(
        ...,
        description="The path to reconstruct the new reference based on existing references.",
    )
    multivalued: bool = Field(
        ..., description="Whether the new reference will be multivalued."
    )


class InferredReference(ReferenceDetails):
    """A model for describing an inferred reference that is based on existing
    references."""

    source: str = Field(
        ..., description="The source class to which this reference should be added."
    )
    target: str = Field(..., description="The class targeted by this reference.")
    new_slot: str = Field(
        ...,
        description="The name of the new slot in the source to store the inferred reference.",
    )

    @root_validator(pre=False)
    @classmethod
    def validate_source_and_target(cls, values):
        """Validate that the source and target attributes are identical with the
        source and target specified in the path."""

        if values.get("source") != values.get("path").source:
            raise ValueError(
                "The source is not identical with the source of the specified path."
            )

        if values.get("target") != values.get("path").target:
            raise ValueError(
                "The target is not identical with the target of the specified path."
            )

        return values

    class Config:
        """Config for this model."""

        frozen = True



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import Generator

import pytest

from metldata.config import SubmissionConfig
from tests.fixtures.metadata_models import VALID_MINIMAL_MODEL_EXAMPLE_PATH

PREFIX_MAPPING = {
    "File": "GHGAF",
    "Experiment": "GHGAE",
    "Sample": "GHGAS",
    "Dataset": "GHGAD",
}


@pytest.fixture
def config_sub_fixture() -> Generator[SubmissionConfig, None, None]:
    """Generate a test config."""

    with TemporaryDirectory() as submission_store_dir:
        with NamedTemporaryFile() as accession_store_path:
            yield SubmissionConfig(
                metadata_model_path=VALID_MINIMAL_MODEL_EXAMPLE_PATH,
                submission_store_dir=submission_store_dir,
                accession_store_path=accession_store_path.name,
                prefix_mapping=PREFIX_MAPPING,
            )



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""API for loading artifacts into running services."""

from typing import Callable, Coroutine, Optional

from fastapi import Depends, HTTPException, Response, Security
from fastapi.routing import APIRouter
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
from ghga_service_commons.auth.context import AuthContextProtocol
from ghga_service_commons.auth.policies import require_auth_context_using_credentials
from hexkit.protocols.dao import DaoFactoryProtocol
from pydantic import BaseModel, Field

from metldata.artifacts_rest.artifact_dao import ArtifactDaoCollection
from metldata.artifacts_rest.artifact_info import get_artifact_info_dict
from metldata.artifacts_rest.models import ArtifactInfo
from metldata.load.auth import check_token
from metldata.load.load import (
    ArtifactResourcesInvalid,
    check_artifact_resources,
    load_artifacts_using_dao,
)
from metldata.load.models import ArtifactResourceDict


class LoaderTokenAuthContext(BaseModel):
    """An auth context used to contain the loader token."""

    token: str = Field(
        ...,
        description="A simple alphanumeric token to authenticate for loading artifacts.",
    )


class LoaderTokenAuthProvider(AuthContextProtocol[LoaderTokenAuthContext]):
    """A provider for the loader token auth context."""

    def __init__(self, *, token_hashes: list[str]):
        """Initialize the loader token auth provider."""

        self._token_hashes = token_hashes

    async def get_context(self, token: str) -> Optional[LoaderTokenAuthContext]:
        """Get a loader token auth context."""

        if not token:
            return None

        if not check_token(token=token, token_hashes=self._token_hashes):
            raise self.AuthContextValidationError("Invalid token.")

        return LoaderTokenAuthContext(token=token)


async def rest_api_factory(
    *,
    artifact_infos: list[ArtifactInfo],
    dao_factory: DaoFactoryProtocol,
    token_hashes: list[str],
    clear_database: Callable[[], Coroutine[None, None, None]]
) -> APIRouter:
    """Return a router for an API for loading artifacts."""

    artifact_info_dict = get_artifact_info_dict(artifact_infos=artifact_infos)
    dao_collection = await ArtifactDaoCollection.construct(
        dao_factory=dao_factory, artifact_infos=artifact_infos
    )

    async def require_token_context(
        credentials: HTTPAuthorizationCredentials = Depends(
            HTTPBearer(auto_error=True)
        ),
    ) -> LoaderTokenAuthContext:
        """Require a VIP authentication and authorization context using FastAPI."""
        return await require_auth_context_using_credentials(
            credentials=credentials,
            auth_provider=LoaderTokenAuthProvider(token_hashes=token_hashes),
        )

    require_token = Security(require_token_context)

    router = APIRouter()

    @router.post("/rpc/load-artifacts")
    async def load_artifacts(
        artifact_resources: ArtifactResourceDict,
        _token: LoaderTokenAuthContext = require_token,
    ):
        """Load artifacts into services for querying via user-accessible API."""

        try:
            check_artifact_resources(
                artifact_resources=artifact_resources, artifact_infos=artifact_info_dict
            )
        except ArtifactResourcesInvalid as error:
            raise HTTPException(status_code=422, detail=str(error)) from error

        await clear_database()

        await load_artifacts_using_dao(
            artifact_resources=artifact_resources,
            artifact_info_dict=artifact_info_dict,
            dao_collection=dao_collection,
        )

        return Response(status_code=204)

    return router



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test the main modules."""

import pytest
from ghga_service_commons.api.testing import AsyncTestClient
from hexkit.protocols.dao import ResourceNotFoundError
from hexkit.providers.mongodb.testutils import mongodb_fixture  # noqa: F401
from hexkit.providers.mongodb.testutils import MongoDbFixture

from metldata.artifacts_rest.artifact_dao import ArtifactDaoCollection
from metldata.artifacts_rest.models import ArtifactInfo
from metldata.load.auth import generate_token, generate_token_and_hash
from metldata.load.config import ArtifactLoaderAPIConfig
from metldata.load.main import get_app
from tests.fixtures.artifact_info import EXAMPLE_ARTIFACT_INFOS
from tests.fixtures.workflows import EXAMPLE_ARTIFACTS


async def get_configured_client(
    mongodb_fixture: MongoDbFixture,  # noqa: F811
    artifact_infos: list[ArtifactInfo],
) -> tuple[AsyncTestClient, str]:
    """Get a tuple of a configured test client together with a corresponding token."""

    token, token_hash = generate_token_and_hash()
    config = ArtifactLoaderAPIConfig(
        artifact_infos=artifact_infos,
        loader_token_hashes=[token_hash],
        **mongodb_fixture.config.dict(),
    )
    app = await get_app(config=config)
    client = AsyncTestClient(app)
    return client, token


@pytest.mark.asyncio
async def test_load_artifacts_endpoint_happy(
    mongodb_fixture: MongoDbFixture,  # noqa: F811
):
    """Test the happy path of using the load artifacts endpoint."""

    client, token = await get_configured_client(
        mongodb_fixture=mongodb_fixture, artifact_infos=EXAMPLE_ARTIFACT_INFOS
    )

    # load example artifacts resources:
    artifact_resources = {
        artifact_name: [artifact_content]
        for artifact_name, artifact_content in EXAMPLE_ARTIFACTS.items()
    }
    response = await client.post(
        "/rpc/load-artifacts",
        json=artifact_resources,
        headers={"Authorization": f"Bearer {token}"},
    )
    assert response.status_code == 204

    # check that the artifact resources were loaded based on an example:
    expected_artifact_name = "inferred_and_public"
    expected_resource_class = "File"
    expected_resource_id = "test_sample_01_R1"
    expected_resource_content = {
        "alias": "test_sample_01_R1",
        "format": "fastq",
        "size": 299943,
    }

    dao_collection = await ArtifactDaoCollection.construct(
        dao_factory=mongodb_fixture.dao_factory, artifact_infos=EXAMPLE_ARTIFACT_INFOS
    )
    dao = await dao_collection.get_dao(
        artifact_name=expected_artifact_name, class_name=expected_resource_class
    )

    observed_resource = await dao.get_by_id(expected_resource_id)
    assert observed_resource.content == expected_resource_content

    # resubmit an empty request:
    response = await client.post(
        "/rpc/load-artifacts",
        json={},
        headers={"Authorization": f"Bearer {token}"},
    )
    assert response.status_code == 204

    # confirm that example resource was deleted:
    with pytest.raises(ResourceNotFoundError):
        await dao.get_by_id(expected_resource_id)


@pytest.mark.asyncio
async def test_load_artifacts_endpoint_invalid_resources(
    mongodb_fixture: MongoDbFixture,  # noqa: F811
):
    """Test using the load artifacts endpoint with resources of unknown artifacts."""

    client, token = await get_configured_client(
        mongodb_fixture=mongodb_fixture, artifact_infos=EXAMPLE_ARTIFACT_INFOS
    )

    # load example artifacts resources:
    unknown_artifact_resources = {
        "unkown_artifact": [list(EXAMPLE_ARTIFACTS.values())[0]]
    }
    response = await client.post(
        "/rpc/load-artifacts",
        json=unknown_artifact_resources,
        headers={"Authorization": f"Bearer {token}"},
    )
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_load_artifacts_endpoint_invalid_token(
    mongodb_fixture: MongoDbFixture,  # noqa: F811
):
    """Test that using the load artifacts endpoint with an invalid token fails."""

    invalid_token = generate_token()
    client, _ = await get_configured_client(
        mongodb_fixture=mongodb_fixture, artifact_infos=EXAMPLE_ARTIFACT_INFOS
    )

    # load artifact resources with invalid token:
    response = await client.post(
        "/rpc/load-artifacts",
        json={},
        headers={"Authorization": f"Bearer {invalid_token}"},
    )
    assert response.status_code == 403



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for transforming metadata."""

from typing import Union, cast

from metldata.builtin_transformations.custom_embeddings.embedding_profile import (
    EmbeddingProfile,
)
from metldata.builtin_transformations.custom_embeddings.model_transform import (
    get_embedding_profile_root_slot,
)
from metldata.metadata_utils import (
    MetadataResourceNotFoundError,
    get_resource_dict_of_class,
    lookup_resource_by_identifier,
    upsert_resources_in_metadata,
)
from metldata.model_utils.anchors import AnchorPoint, lookup_anchor_point
from metldata.model_utils.essentials import MetadataModel
from metldata.transform.base import Json, MetadataTransformationError


def is_slot_multivalued(
    *, slot_name: str, class_name: str, model: MetadataModel
) -> bool:
    """Checks whether a slot is multivalued."""

    slot_definition = model.schema_view.induced_slot(
        slot_name=slot_name, class_name=class_name
    )

    if not slot_definition:
        raise RuntimeError(  # this should never happen
            f"Slot '{slot_name}' not found in class '{class_name}'"
        )

    return bool(slot_definition.multivalued)


def resolve_target_resource(
    target_resource_id: str,
    target: Union[str, EmbeddingProfile],
    global_metadata: Json,
    model: MetadataModel,
    anchor_points_by_target: dict,
) -> Json:
    """Resolves a target resource.

    Raises:
        MetadataTransformationError: If the target resource could not be found.
    """

    if isinstance(target, EmbeddingProfile):
        return generate_embedded_resource(
            resource_id=target_resource_id,
            embedding_profile=target,
            global_metadata=global_metadata,
            model=model,
            anchor_points_by_target=anchor_points_by_target,
        )

    try:
        return lookup_resource_by_identifier(
            class_name=target,
            identifier=target_resource_id,
            global_metadata=global_metadata,
            anchor_points_by_target=anchor_points_by_target,
        )
    except MetadataResourceNotFoundError as error:
        raise MetadataTransformationError(
            f"Could not find resource '{target_resource_id}' of class '{target}'"
        ) from error


def generate_embedded_resource(
    resource_id: str,
    embedding_profile: EmbeddingProfile,
    global_metadata: Json,
    model: MetadataModel,
    anchor_points_by_target: dict,
) -> Json:
    """Generates an embedded version of the specified resource. This is done recursively
    for all embedded references that are linked to this resource."""

    resource = lookup_resource_by_identifier(
        class_name=embedding_profile.source_class,
        identifier=resource_id,
        global_metadata=global_metadata,
        anchor_points_by_target=anchor_points_by_target,
    )

    for reference_slot_name, target in embedding_profile.embedded_references.items():
        if is_slot_multivalued(
            slot_name=reference_slot_name,
            class_name=embedding_profile.source_class,
            model=model,
        ):
            target_resource_ids = cast(list[str], resource[reference_slot_name])
            target_resources = [
                resolve_target_resource(
                    target_resource_id=target_resource_id,
                    target=target,
                    global_metadata=global_metadata,
                    model=model,
                    anchor_points_by_target=anchor_points_by_target,
                )
                for target_resource_id in target_resource_ids
            ]
            resource[reference_slot_name] = target_resources
        else:
            target_resource_id = cast(str, resource[reference_slot_name])
            target_resource = resolve_target_resource(
                target_resource_id=target_resource_id,
                target=target,
                global_metadata=global_metadata,
                model=model,
                anchor_points_by_target=anchor_points_by_target,
            )
            resource[reference_slot_name] = target_resource

    return resource


def add_custom_embedding_to_metadata(
    *,
    metadata: Json,
    embedding_profile: EmbeddingProfile,
    model: MetadataModel,
    anchor_points_by_target: dict[str, AnchorPoint],
) -> Json:
    """Add custom embedding to the metadata.

    Raises:
        MetadataTransformationError:
            if the transformation of the metadata fails.
    """

    resource_ids = get_resource_dict_of_class(
        class_name=embedding_profile.source_class,
        global_metadata=metadata,
        anchor_points_by_target=anchor_points_by_target,
    ).keys()

    resources = [
        generate_embedded_resource(
            resource_id=resource_id,
            embedding_profile=embedding_profile,
            global_metadata=metadata,
            model=model,
            anchor_points_by_target=anchor_points_by_target,
        )
        for resource_id in resource_ids
    ]

    # add anchor point for embedding profile:
    source_class_anchor_point = lookup_anchor_point(
        class_name=embedding_profile.source_class,
        anchor_points_by_target=anchor_points_by_target,
    )
    embedded_class_anchor_point = AnchorPoint(
        target_class=embedding_profile.target_class,
        identifier_slot=source_class_anchor_point.identifier_slot,
        root_slot=get_embedding_profile_root_slot(embedding_profile=embedding_profile),
    )
    anchor_points_by_target_modified = anchor_points_by_target.copy()
    anchor_points_by_target_modified[
        embedding_profile.target_class
    ] = embedded_class_anchor_point

    return upsert_resources_in_metadata(
        resources=resources,
        class_name=embedding_profile.target_class,
        global_metadata=metadata,
        anchor_points_by_target=anchor_points_by_target_modified,
    )


def add_custom_embeddings_to_metadata(
    *,
    metadata: Json,
    embedding_profiles: list[EmbeddingProfile],
    model: MetadataModel,
    anchor_points_by_target: dict[str, AnchorPoint],
) -> Json:
    """Add custom embeddings to the metadata.

    Raises:
        MetadataTransformationError:
            if the transformation of the metadata fails.
    """

    for embedding_profile in embedding_profiles:
        metadata = add_custom_embedding_to_metadata(
            metadata=metadata,
            embedding_profile=embedding_profile,
            model=model,
            anchor_points_by_target=anchor_points_by_target,
        )

    return metadata



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Handling of the accession handler."""

import secrets
from typing import Iterable

from pydantic import BaseSettings, Field

from metldata.accession_registry.accession_store import AccessionStore


class AccessionRegistryConfig(BaseSettings):
    """Config parameters and their defaults."""

    prefix_mapping: dict[str, str] = Field(
        ...,
        description="Specifies the ID prefix (values) per resource type (keys).",
        example={
            "file": "GHGAF",
            "experiment": "GHGAE",
            "sample": "GHGAS",
            "dataset": "GHGAD",
        },
    )

    suffix_length: int = Field(8, description="Length of the numeric ID suffix.")


class AccessionRegistry:
    """Main class handling the accession registry."""

    class UnkownResourceTypeError(RuntimeError):
        """Raised when a resource type is specified that is unkown."""

        def __init__(self, *, specified_type: str, expected_types: Iterable[str]):
            """Specify the given as well as the expected resource types."""

            message = (
                f"The specified resource type '{specified_type} is unkown'."
                + " Expected one of: "
                + ", ".join(expected_types)
            )
            super().__init__(message)

    class AccessionGenerationError(RuntimeError):
        """Raised when a the generation of a new accession failed."""

    def __init__(
        self, *, config: AccessionRegistryConfig, accession_store: AccessionStore
    ):
        """Initialize with config."""

        self._config = config
        self._accession_store = accession_store

    def _assert_resource_type_exists(self, *, resource_type: str) -> None:
        """Checks whether the specified resource type is in the prefix mapping, raises
        and UnkownResourceTypeError otherwise."""

        if resource_type not in self._config.prefix_mapping:
            raise self.UnkownResourceTypeError(
                specified_type=resource_type,
                expected_types=self._config.prefix_mapping.keys(),
            )

    def _generate_accession(self, *, resource_type: str) -> str:
        """Generate a new accession."""

        self._assert_resource_type_exists(resource_type=resource_type)

        prefix = self._config.prefix_mapping[resource_type]
        suffix = "".join(
            str(secrets.randbelow(10)) for _ in range(self._config.suffix_length)
        )

        return prefix + suffix

    def get_accession(self, *, resource_type: str) -> str:
        """
        Generates and registers a new accession for a resource of the specified type.
        """

        for _ in range(10):
            # try 10 times to generate a new accession:
            accession = self._generate_accession(resource_type=resource_type)

            try:
                self._accession_store.save(accession=accession)
            except AccessionStore.AccessionAlreadyExistsError:
                continue

            return accession

        raise self.AccessionGenerationError(
            "Tried and failed 10 times to generate a new accession that is not used"
            + " already. The accession space might be exhausted."
        )



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Storing and exploring existing accessions."""

from pathlib import Path

from pydantic import BaseSettings, Field


class AccessionStoreConfig(BaseSettings):
    """Config parameters and their defaults."""

    accession_store_path: Path = Field(
        ..., description="A file for storing the already registered accessions."
    )


class AccessionStore:
    """A class for storing and querying existing accessions."""

    class AccessionAlreadyExistsError(RuntimeError):
        """Raised when an accession already exists."""

        def __init__(self, *, accession: str):
            message = f"The following accession already exists: {accession}"
            super().__init__(message)

    def __init__(self, *, config: AccessionStoreConfig):
        """Initialize with config parameters."""

        self._config = config

    def exists(self, *, accession: str) -> bool:
        """Checks whether the given accession is already in use."""

        with open(self._config.accession_store_path, "r", encoding="utf-8") as store:
            for existing_accession in store:
                if accession == existing_accession.strip():
                    return True

        return False

    def save(self, *, accession: str) -> None:
        """Save a new accession.

        Raises:
            AccessionAlreadyExistsError: If the given accession already exists.
        """

        if self.exists(accession=accession):
            raise self.AccessionAlreadyExistsError(accession=accession)

        with open(self._config.accession_store_path, "a", encoding="utf-8") as store:
            store.write(f"{accession}\n")



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Transformation test cases."""

from dataclasses import dataclass
from typing import Generic, TypeVar

from pydantic import BaseModel

from metldata.builtin_transformations.add_accessions import (
    ACCESSION_ADDITION_TRANSFORMATION,
)
from metldata.builtin_transformations.custom_embeddings import (
    CUSTOM_EMBEDDING_TRANSFORMATION,
)
from metldata.builtin_transformations.delete_slots import SLOT_DELETION_TRANSFORMATION
from metldata.builtin_transformations.infer_references import (
    REFERENCE_INFERENCE_TRANSFORMATION,
)
from metldata.builtin_transformations.merge_slots import SLOT_MERGING_TRANSFORMATION
from metldata.custom_types import Json
from metldata.event_handling.models import SubmissionAnnotation
from metldata.model_utils.essentials import MetadataModel
from metldata.transform.base import TransformationDefinition
from tests.fixtures.utils import BASE_DIR, read_yaml

Config = TypeVar("Config", bound=BaseModel)


@dataclass(frozen=True)
class TransformationTestCase(Generic[Config]):
    """A test case for a transformation."""

    transformation_name: str
    case_name: str
    transformation_definition: TransformationDefinition
    config: Config
    original_model: MetadataModel
    original_metadata: Json
    metadata_annotation: SubmissionAnnotation
    transformed_model: MetadataModel
    transformed_metadata: Json

    def __str__(self) -> str:
        return f"{self.transformation_name}-{self.case_name}"


def _read_test_case(
    *,
    transformation_name: str,
    transformation_definition: TransformationDefinition,
    case_name: str,
) -> TransformationTestCase[Config]:
    """Read a test case for a transformation."""

    case_dir = BASE_DIR / "transformations" / transformation_name / case_name
    config_path = case_dir / "config.yaml"
    original_model_path = case_dir / "original_model.yaml"
    original_metadata_path = case_dir / "original_metadata.yaml"
    metadata_annotation_path = case_dir / "metadata_annotation.yaml"
    transformed_model_path = case_dir / "transformed_model.yaml"
    transformed_metadata_path = case_dir / "transformed_metadata.yaml"

    return TransformationTestCase(
        transformation_name=transformation_name,
        case_name=case_name,
        transformation_definition=transformation_definition,
        config=transformation_definition.config_cls(**read_yaml(config_path)),
        original_model=MetadataModel.init_from_path(original_model_path),
        original_metadata=read_yaml(original_metadata_path),
        metadata_annotation=(
            SubmissionAnnotation(**read_yaml(metadata_annotation_path))
            if metadata_annotation_path.exists()
            else SubmissionAnnotation(accession_map={})
        ),
        transformed_model=MetadataModel.init_from_path(transformed_model_path),
        transformed_metadata=read_yaml(transformed_metadata_path),
    )


def _read_all_test_cases_for_a_transformation(
    *,
    transformation_name: str,
    transformation_definition: TransformationDefinition,
) -> list[TransformationTestCase]:
    """Read all test cases for a transformation."""

    base_dir = BASE_DIR / "transformations" / transformation_name
    case_names = [path.name for path in base_dir.iterdir() if path.is_dir()]

    return [
        _read_test_case(
            transformation_name=transformation_name,
            transformation_definition=transformation_definition,
            case_name=case_name,
        )
        for case_name in case_names
    ]


def _read_all_test_cases(
    *, transformations_by_name: dict[str, TransformationDefinition]
) -> list[TransformationTestCase]:
    """Read all test cases for the specified transformations."""

    return [
        test_case
        for transformation_name, transformation_definition in transformations_by_name.items()
        for test_case in _read_all_test_cases_for_a_transformation(
            transformation_name=transformation_name,
            transformation_definition=transformation_definition,
        )
    ]


TRANSFORMATIONS_BY_NAME: dict[str, TransformationDefinition] = {
    "add_accessions": ACCESSION_ADDITION_TRANSFORMATION,
    "infer_references": REFERENCE_INFERENCE_TRANSFORMATION,
    "delete_slots": SLOT_DELETION_TRANSFORMATION,
    "custom_embedding": CUSTOM_EMBEDDING_TRANSFORMATION,
    "merge_slots": SLOT_MERGING_TRANSFORMATION,
}


TRANSFORMATION_TEST_CASES = _read_all_test_cases(
    transformations_by_name=TRANSFORMATIONS_BY_NAME
)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test the identifiers module."""

from metldata.model_utils.identifiers import get_class_identifiers
from tests.fixtures.metadata_models import VALID_MINIMAL_METADATA_MODEL


def test_get_class_identifiers_happy():
    """Test the happy path of using the get_class_identifiers function."""

    expected_identifiers = {"File": "alias", "Dataset": "alias"}

    observed_identifiers = get_class_identifiers(model=VALID_MINIMAL_METADATA_MODEL)

    assert observed_identifiers == expected_identifiers



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic to check basic assumptions about the metadata model."""
from typing import Optional

from metldata.model_utils.anchors import (
    AnchorPointNotFoundError,
    InvalidAnchorPointError,
    filter_anchor_points,
    get_anchors_points_by_target,
)
from metldata.model_utils.essentials import ROOT_CLASS, MetadataModel


class MetadataModelAssumptionError(RuntimeError):
    """Raised when the assumptions about the metadata model are not met."""


def check_class_exists(model: MetadataModel, class_name: str) -> None:
    """Check that a class with the given name exists.

    Raises:
        MetadataModelAssumptionError: if validation fails.
    """

    # has a class called class_name:
    class_ = model.schema_view.get_class(class_name=class_name)

    if class_ is None:
        raise MetadataModelAssumptionError(
            f"A class called '{class_name}' is required but does not exist."
        )


def check_class_slot_exists(
    model: MetadataModel, class_name: str, slot_name: str, ignore_parents: bool = False
) -> None:
    """Check that a class with the given name exists and that it has a slot with the
    given name. If ignore_parents is set to True, slots that are inherited from parent
    classes or mixins are ignored.

    Raises:
        MetadataModelAssumptionError: if validation fails.
    """

    check_class_exists(model=model, class_name=class_name)

    all_slots = model.schema_view.class_slots(
        class_name=class_name, direct=ignore_parents
    )

    if slot_name not in all_slots:
        raise MetadataModelAssumptionError(
            f"A slot called '{slot_name}' is required but does not exist"
            + f" in the '{class_name}' class."
            " Inherited slots are ignored."
            if ignore_parents
            else f"A slot called '{slot_name}' is required but does not exist"
            + f" in the '{class_name}' class or its parents classes and mixins."
        )


def check_root_class_existence(model: MetadataModel) -> None:
    """Check the existence of a root class that is called as defined in ROOT_CLASS.

    Raises:
        MetadataModelAssumptionError: if validation fails.
    """

    # has a tree root called ROOT_CLASS:
    root_class = model.schema_view.get_class(class_name=ROOT_CLASS)

    if root_class is None:
        raise MetadataModelAssumptionError(
            f"A {ROOT_CLASS} class is required but does not exist."
        )
    if not root_class.tree_root:
        raise MetadataModelAssumptionError(
            f"The {ROOT_CLASS} class must have the tree_root property set to true."
        )


def check_anchor_points(
    model: MetadataModel, classes: Optional[list[str]] = None
) -> None:
    """Checks the anchor points of the root class. If classes is specified, also checks
    that anchor points for the classes exist.

    Raises:
        MetadataModelAssumptionError: if validation fails.
    """

    try:
        anchor_points_by_target = get_anchors_points_by_target(model=model)
    except InvalidAnchorPointError as error:
        raise MetadataModelAssumptionError(
            f"The model has invalid anchor points: {error}"
        ) from error

    if classes is not None:
        try:
            _ = filter_anchor_points(
                anchor_points_by_target=anchor_points_by_target,
                classes_of_interest=set(classes),
            )
        except AnchorPointNotFoundError as error:
            raise MetadataModelAssumptionError(
                f"The model is missing anchor points: {error}"
            ) from error


def check_basic_model_assumption(model: MetadataModel) -> None:
    """Check that the basic assumptions that metldata makes about the metadata model
    are met. Raises a MetadataModelAssumptionError otherwise.

    Raises:
        MetadataModelAssumptionError: if validation fails.
    """

    check_root_class_existence(model=model)
    check_anchor_points(model=model)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Config parameters and their defaults."""

from pydantic import BaseSettings, Field, validator

from metldata.builtin_transformations.merge_slots.models import SlotMergeInstruction


class SlotMergingConfig(BaseSettings):
    """Config containing slots to be deleted from models and associated metadata."""

    merge_instructions: list[SlotMergeInstruction] = Field(
        ...,
        description=(
            "A list of slot merging instructions. Each instruction specifies a class"
            + " and a target slot into which the source slots should be merged."
            + " You may specify merge instructions for the same class."
            + " However, the target slot of one merge instruction cannot be used as"
            + " a source slot in another merge instruction."
            + " The source slots will not be deleted."
        ),
        example=[
            {
                "class_name": "class_a",
                "source_slots": ["some_slot", "another_slot"],
                "target_slot": "merged_slot",
            },
        ],
    )

    # pylint: disable=no-self-argument
    @validator("merge_instructions")
    def validate_merge_instructions(
        cls, filtered_merge_instructions: list[SlotMergeInstruction]
    ) -> list[SlotMergeInstruction]:
        """Validate that source and target slots do not overlap across merge
        instructions and that no target slot is reused for the same
        class."""

        class_names = {merge.class_name for merge in filtered_merge_instructions}
        for class_name in class_names:
            filtered_merge_instructions = [
                merge
                for merge in filtered_merge_instructions
                if merge.class_name == class_name
            ]
            source_slots = {
                slot
                for merge in filtered_merge_instructions
                for slot in merge.source_slots
            }

            target_slot_list = [
                merge.target_slot for merge in filtered_merge_instructions
            ]
            target_slots = set(target_slot_list)

            if len(target_slot_list) != len(target_slots):
                raise ValueError(
                    f"Multiple merge instructions for class '{class_name}'"
                    + " have the same target slot."
                )

            if source_slots.intersection(target_slots):
                raise ValueError(
                    f"Source and target slots for class '{class_name}' overlap."
                )

        return filtered_merge_instructions

    class Config:
        """Pydantic config."""

        extra = "forbid"



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Handling identifiers of classes in a metadata model."""

from typing import Optional

from metldata.model_utils.essentials import ROOT_CLASS, MetadataModel


def get_class_identifier(model: MetadataModel, class_name: str) -> Optional[str]:
    """Get the identifier of a class in a metadata model.

    Returns:
        The identifier of the class, or None if the class does not have an identifier.
    """

    schema_view = model.schema_view

    for slot_name in schema_view.class_slots(class_name=class_name):
        slot_def = schema_view.induced_slot(slot_name=slot_name, class_name=class_name)
        if slot_def.identifier:
            return str(slot_def.name)

    return None


def get_class_identifiers(model: MetadataModel) -> dict[str, Optional[str]]:
    """Get the identifiers of all classes in a metadata model.

    Returns:
        A dictionary with the class names as keys and the identifiers as values. If a
        class does not have an identifier, the value is None.
    """

    schema_view = model.schema_view

    identifiers_by_class: dict[str, Optional[str]] = {}
    for class_name in schema_view.all_classes():
        if class_name == ROOT_CLASS:
            continue  # Root class does not have an identifier
        class_def = schema_view.get_class(class_name=class_name)
        if class_def.mixin or class_def.abstract:
            continue  # Mixins and abstract classes do not have an identifier
        identifier = get_class_identifier(model=model, class_name=class_name)
        identifiers_by_class[class_name] = identifier

    return identifiers_by_class



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for querying artifact resouces."""

from hexkit.protocols.dao import ResourceNotFoundError

from metldata.artifacts_rest.artifact_dao import ArtifactDaoCollection
from metldata.artifacts_rest.models import ArtifactResource


class ArtifactResourceNotFoundError(RuntimeError):
    """Raised when a resource could not be found."""

    def __init__(self, *, artifact_name: str, class_name: str, resource_id: str):
        message = (
            f"Could not find resource with ID '{resource_id}' of class '{class_name}'"
            + f" in artifact '{artifact_name}'."
        )
        super().__init__(message)


async def query_artifact_resource(
    *,
    artifact_name: str,
    class_name: str,
    resource_id: str,
    dao_collection: ArtifactDaoCollection,
) -> ArtifactResource:
    """Query a single resource with the given ID of the given class from the given
    artifact.

    Raises:
        ArtifactResourceNotFoundError: If the resource could not be found.
    """

    dao = await dao_collection.get_dao(
        artifact_name=artifact_name, class_name=class_name
    )

    try:
        return await dao.get_by_id(resource_id)
    except ResourceNotFoundError as error:
        raise ArtifactResourceNotFoundError(
            artifact_name=artifact_name,
            class_name=class_name,
            resource_id=resource_id,
        ) from error



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for handling identifiers and accessions."""

from functools import partial
from typing import Iterable, Optional
from uuid import uuid4

from pydantic import Json

from metldata.accession_registry import AccessionRegistry
from metldata.custom_types import SubmissionContent
from metldata.metadata_utils import lookup_self_id
from metldata.model_utils.anchors import (
    AnchorPoint,
    invert_anchor_points_by_target,
    lookup_anchor_point,
    lookup_class_by_anchor_point,
)
from metldata.submission_registry.models import AccessionMap


def generate_submission_id() -> str:
    """Generate a new submission ID."""

    return str(uuid4())


def lookup_accession(
    *,
    anchor: str,
    alias: str,
    accession_map: AccessionMap,
) -> Optional[str]:
    """Lookup the accession for a resource with the provided user-defined alias of
    the class with the provided anchor. If no accession exists, None is returned."""

    return accession_map.get(anchor, {}).get(alias, None)


def get_accession(
    *,
    anchor: str,
    alias: str,
    accession_map: AccessionMap,
    accession_registry: AccessionRegistry,
    target_by_anchor_point: dict[str, str],
) -> str:
    """Get the accession for a resource with the provided user-defined alias of
    the class with the provided anchor. If no entry can be found in the provided accession map,
    a new accession is requested from the provided accession registry."""

    accession = lookup_accession(
        anchor=anchor, alias=alias, accession_map=accession_map
    )
    if accession:
        return accession

    class_name = lookup_class_by_anchor_point(
        root_slot=anchor, target_by_anchor_point=target_by_anchor_point
    )
    return accession_registry.get_accession(resource_type=class_name)


def get_aliases_for_resources(
    *,
    resources: list[Json],
    root_slot: str,
    anchor_points_by_target: dict[str, AnchorPoint],
) -> Iterable[str]:
    """Get the aliases for the provided resources."""

    target_by_anchor_point = invert_anchor_points_by_target(
        anchor_points_by_target=anchor_points_by_target
    )

    class_name = lookup_class_by_anchor_point(
        root_slot=root_slot, target_by_anchor_point=target_by_anchor_point
    )
    anchor_point = lookup_anchor_point(
        class_name=class_name, anchor_points_by_target=anchor_points_by_target
    )

    for resource in resources:
        yield lookup_self_id(
            resource=resource, identifier_slot=anchor_point.identifier_slot
        )


def generate_accession_map(
    *,
    content: SubmissionContent,
    existing_accession_map: Optional[AccessionMap] = None,
    accession_registry: AccessionRegistry,
    anchor_points_by_target: dict[str, AnchorPoint],
) -> AccessionMap:
    """Generate an accession map for the provided content.

    If an existing accession map is provided, an updated version is returned. Thereby,
    new accessions are added for new content resources. Existing accessions are kept if
    the corresponding content resource still exists, otherwise they are removed."""

    if existing_accession_map is None:
        existing_accession_map = {}

    target_by_anchor_point = invert_anchor_points_by_target(
        anchor_points_by_target=anchor_points_by_target
    )

    get_accession_ = partial(
        get_accession,
        accession_map=existing_accession_map,
        accession_registry=accession_registry,
        target_by_anchor_point=target_by_anchor_point,
    )

    return {
        anchor: {
            alias: get_accession_(anchor=anchor, alias=alias)
            for alias in get_aliases_for_resources(
                resources=resources,
                root_slot=anchor,
                anchor_points_by_target=anchor_points_by_target,
            )
        }
        for anchor, resources in content.items()
    }



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Testing the metadata validator."""

from contextlib import nullcontext
from typing import Any

import pytest

from metldata.model_utils.metadata_validator import MetadataValidator
from tests.fixtures.metadata import (
    INVALID_MINIMAL_METADATA_EXAMPLES,
    VALID_MINIMAL_METADATA_EXAMPLES,
)
from tests.fixtures.metadata_models import VALID_MINIMAL_METADATA_MODEL


@pytest.mark.parametrize(
    "metadata, is_valid",
    [(valid_metadata, True) for valid_metadata in VALID_MINIMAL_METADATA_EXAMPLES]
    + [
        (invalid_metadata, False)
        for invalid_metadata in INVALID_MINIMAL_METADATA_EXAMPLES
    ],
)
def test_validate_against_model(metadata: dict[str, Any], is_valid: bool):
    """Test the validation of metadata against a model."""

    validator = MetadataValidator(model=VALID_MINIMAL_METADATA_MODEL)

    with nullcontext() if is_valid else pytest.raises(  # type:ignore
        MetadataValidator.ValidationError
    ):
        validator.validate(metadata)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Testing the assumptions module."""

from contextlib import nullcontext

import pytest

from metldata.model_utils.assumptions import (
    MetadataModelAssumptionError,
    check_basic_model_assumption,
    check_class_exists,
    check_class_slot_exists,
)
from metldata.model_utils.essentials import MetadataModel
from tests.fixtures.metadata_models import (
    INVALID_METADATA_MODELS,
    VALID_METADATA_MODELS,
    VALID_MINIMAL_METADATA_MODEL,
)


@pytest.mark.parametrize(
    "model, is_valid",
    [(valid_model, True) for valid_model in VALID_METADATA_MODELS]
    + [(invalid_model, False) for invalid_model in INVALID_METADATA_MODELS],
)
def test_metadata_model_assumption_checking(model: MetadataModel, is_valid: bool):
    """Test the assumptions regarding the metadata model are correctly checked."""

    with nullcontext() if is_valid else pytest.raises(  # type:ignore
        MetadataModelAssumptionError
    ):
        check_basic_model_assumption(model)


@pytest.mark.parametrize(
    "model, class_name, exists",
    [
        (VALID_MINIMAL_METADATA_MODEL, "Dataset", True),
        (VALID_MINIMAL_METADATA_MODEL, "NotExistingClass", False),
    ],
)
def test_check_class_exists(model: MetadataModel, class_name: str, exists: bool):
    """Test the check_class_exists function."""

    with nullcontext() if exists else pytest.raises(  # type:ignore
        MetadataModelAssumptionError
    ):
        check_class_exists(model=model, class_name=class_name)


@pytest.mark.parametrize(
    "model, class_name, slot_name, ignore_parents, exists",
    [
        (VALID_MINIMAL_METADATA_MODEL, "Dataset", "alias", False, True),
        (VALID_MINIMAL_METADATA_MODEL, "Dataset", "alias", True, False),
        (VALID_MINIMAL_METADATA_MODEL, "Dataset", "files", True, True),
        (VALID_MINIMAL_METADATA_MODEL, "Dataset", "not_existing_slot", False, False),
        (VALID_MINIMAL_METADATA_MODEL, "NotExistingClass", "name", False, False),
    ],
)
def test_check_class_slot_exists(
    model: MetadataModel,
    class_name: str,
    slot_name: str,
    ignore_parents: bool,
    exists: bool,
):
    """Test the check_class_slot_exists function."""

    with nullcontext() if exists else pytest.raises(  # type:ignore
        MetadataModelAssumptionError
    ):
        check_class_slot_exists(
            model=model,
            class_name=class_name,
            slot_name=slot_name,
            ignore_parents=ignore_parents,
        )



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test the query_resource module."""

import pytest
from hexkit.providers.mongodb.testutils import mongodb_fixture  # noqa: F401
from hexkit.providers.mongodb.testutils import MongoDbFixture

from metldata.artifacts_rest.query_resources import query_artifact_resource
from tests.artifact_rest.test_load_artifacts import load_example_artifact_resources
from tests.fixtures.artifact_info import MINIMAL_ARTIFACT_INFO


@pytest.mark.asyncio
async def test_query_artifact_resource(mongodb_fixture: MongoDbFixture):  # noqa: F811
    """Test happy path of using the query_artifact_resource function."""

    # load example resources and prepare client:
    dao_collection = await load_example_artifact_resources(
        dao_factory=mongodb_fixture.dao_factory
    )

    # Get an example resource:
    artifact_name = MINIMAL_ARTIFACT_INFO.name
    class_name = "File"
    resource_id = "test_sample_01_R1"
    observed_resource = await query_artifact_resource(
        artifact_name=artifact_name,
        class_name=class_name,
        resource_id=resource_id,
        dao_collection=dao_collection,
    )

    # check that the right object was return using an example slot:
    expected_checksum = (
        "1c8aed294d5dec3740a175f6b655725fa668bfe41311e74f7ca9d85c91371b4e"
    )
    assert observed_resource.content["checksum"] == expected_checksum
    # check that the ID is included in the resource:
    assert observed_resource.content["alias"] == resource_id



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for transforming metadata models."""


from metldata.model_utils.essentials import MetadataModel
from metldata.model_utils.manipulate import delete_class_slot


def delete_class_slots_from_model(
    model: MetadataModel, class_slots: dict[str, list[str]]
) -> MetadataModel:
    """Delete the specified class slots from the provided model. Returns a modified
    copy of the model."""

    schema_view = model.schema_view

    for class_name, slot_names in class_slots.items():
        for slot_name in slot_names:
            schema_view = delete_class_slot(
                schema_view=schema_view, class_name=class_name, slot_name=slot_name
            )

    return schema_view.export_model()



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for subscribing to source events."""

from typing import Awaitable, Callable

from hexkit.custom_types import Ascii, JsonObject
from hexkit.protocols.eventsub import EventSubscriberProtocol

from metldata.event_handling.models import SubmissionEventPayload
from metldata.event_handling.submission_events import SourceEventConfig


class SourceEventSubscriberConfig(SourceEventConfig):
    """Config parameters and their defaults."""


class SourceEventSubscriber(EventSubscriberProtocol):
    """Consumes source events."""

    def __init__(
        self,
        *,
        config: SourceEventSubscriberConfig,
        run_workflow_func: Callable[[SubmissionEventPayload], Awaitable[None]]
    ):
        """Initialize with config parameters."""

        self.topics_of_interest = [config.source_event_topic]
        self.types_of_interest = [config.source_event_type]
        self._run_workflow_func = run_workflow_func

    # pylint: disable=unused-argument
    async def _consume_validated(
        self, *, payload: JsonObject, type_: Ascii, topic: Ascii
    ) -> None:
        """
        Receive and process an event with already validated topic and type.

        Args:
            payload (JsonObject): The data/payload to send with the event.
            type_ (str): The type of the event.
            topic (str): Name of the topic the event was published to.
        """

        submission_event_payload = SubmissionEventPayload(**payload)
        await self._run_workflow_func(submission_event_payload)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Basic constants and logic related to models."""

from __future__ import annotations

import dataclasses
import json
from contextlib import contextmanager
from copy import copy, deepcopy
from functools import lru_cache
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, Generator

import jsonasobj2
import yaml
from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import SchemaDefinition

# The name of the root class of a model:
ROOT_CLASS = "Submission"


@lru_cache
def schema_view_from_model(model: MetadataModel) -> SchemaView:
    """Get a schema view instance from the metadata model."""

    return ExportableSchemaView(model)


class MetadataModel(SchemaDefinition):
    """A dataclass for describing metadata models."""

    @classmethod
    def init_from_path(cls, model_path: Path) -> MetadataModel:
        """Initialize from a model file in yaml format."""

        with open(model_path, "r", encoding="utf-8") as file:
            model_json = yaml.safe_load(file)

        return cls(**model_json)

    @property
    def schema_view(self) -> ExportableSchemaView:
        """Get a schema view instance from the metadata model."""

        return schema_view_from_model(self)

    def copy(self) -> MetadataModel:
        """Copy the model."""

        return deepcopy(self)

    def __eq__(self, other: object):
        """For comparisons."""

        if not isinstance(other, MetadataModel):
            return NotImplemented

        return self.as_dict() == other.as_dict()

    # pylint: disable=too-many-nested-blocks,too-many-branches
    def as_dict(self, essential: bool = True):  # noqa: C901
        """Get a dictionary representation of the model. If essential set to True, the
        dictionary will be cleaned of all fields that are not essential."""

        model_dict = dataclasses.asdict(self)

        if essential:
            if "classes" in model_dict:
                for class_ in model_dict["classes"].values():
                    if "from_schema" in class_:
                        del class_["from_schema"]
                    if "slot_usage" in class_:
                        slot_usage: Any = (
                            jsonasobj2.as_dict(class_["slot_usage"])
                            if isinstance(class_["slot_usage"], jsonasobj2.JsonObj)
                            else class_["slot_usage"]
                        )
                        for slot in slot_usage.values():
                            if "alias" in slot:
                                del slot["alias"]
                            if "from_schema" in slot:
                                del slot["from_schema"]
                            if "domain_of" in slot:
                                del slot["domain_of"]
            if "slots" in model_dict:
                for slot in model_dict["slots"].values():
                    if "from_schema" in slot:
                        del slot["from_schema"]
            if "enums" in model_dict:
                for enum in model_dict["enums"].values():
                    if "from_schema" in enum:
                        del enum["from_schema"]

        return model_dict

    def as_json(self) -> str:
        """Get a json representation of the model."""

        return json.dumps(self.as_dict(essential=True), indent=2)

    def as_yaml(self) -> str:
        """Get a yaml representation of the model."""

        return yaml.safe_dump(self.as_dict(essential=True))

    def write_yaml(self, path: Path) -> None:
        """Write the model to a yaml file."""

        with open(path, "w", encoding="utf-8") as file:
            yaml.safe_dump(self.as_dict(essential=True), file)

    @contextmanager
    def temporary_yaml_path(self) -> Generator[Path, None, None]:
        """Returns a context manager that creates a temporary yaml file containing the
        model and returns its path on __enter__ and deletes it on __exit__.

        This is required because some tools in the linkml schema ecosystem only support
        working with paths to yaml file and not with in-memory representations.
        """

        with NamedTemporaryFile(mode="w", encoding="utf-8") as file:
            model_json = self.as_dict()
            yaml.safe_dump(model_json, file)
            file.flush()
            yield Path(file.name)

    def __hash__(self):
        """Return a hash of the model."""

        return hash(self.as_json())


class ExportableSchemaView(SchemaView):
    """Extend the SchemaView by adding a method for exporting a MetadataModel."""

    def __copy__(self):
        """Return a copy of the model.

        Please note, the copy will have a new uuid used for hashing.
        """

        return ExportableSchemaView(
            schema=copy(self.schema),
            importmap=copy(self.importmap),
        )

    def __deepcopy__(self, memo: Any):
        """Return a deepcopy of the model.

        Please note, the copy will have a new uuid used for hashing.
        """

        return ExportableSchemaView(
            schema=deepcopy(self.schema), importmap=copy(self.importmap)
        )

    def export_model(self) -> MetadataModel:
        """Export a MetadataModel."""

        model_json = dataclasses.asdict(deepcopy(self.schema))

        return MetadataModel(**model_json)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for transforming metadata models."""

from copy import deepcopy
from typing import Union

from linkml_runtime.linkml_model.meta import ClassDefinition, SlotDefinition
from stringcase import snakecase

from metldata.builtin_transformations.custom_embeddings.embedding_profile import (
    EmbeddingProfile,
)
from metldata.model_utils.anchors import AnchorPoint, get_anchors_points_by_target
from metldata.model_utils.essentials import (
    ROOT_CLASS,
    ExportableSchemaView,
    MetadataModel,
)
from metldata.model_utils.identifiers import get_class_identifier
from metldata.model_utils.manipulate import (
    add_anchor_point,
    add_slot_usage_annotation,
    disable_identifier_slot,
    get_normalized_slot_usage,
)
from metldata.transform.base import MetadataModelTransformationError


def get_embedding_profile_root_slot(embedding_profile: EmbeddingProfile) -> str:
    """Get the root slot for an embedding profile."""
    return snakecase(embedding_profile.target_class)


def get_embedded_reference_slot(
    *,
    class_: ClassDefinition,
    schema_view: ExportableSchemaView,
    reference_slot_name: str,
    target: Union[str, EmbeddingProfile],
) -> SlotDefinition:
    """Update the slot definition of an embedded reference."""

    if not class_.slots or reference_slot_name not in class_.slots:
        raise MetadataModelTransformationError(
            f"Class '{class_.name}' does not have the slot '{reference_slot_name}'"
        )

    slot_usage = get_normalized_slot_usage(class_=class_)
    if reference_slot_name in slot_usage:
        slot_definition = slot_usage[reference_slot_name]
    else:
        slot_definition = SlotDefinition(name=reference_slot_name)

    # make sure that the range is in line with the embedding profile:
    expected_target_class = (
        target.source_class if isinstance(target, EmbeddingProfile) else target
    )
    induced_slot_definition = schema_view.induced_slot(
        slot_name=reference_slot_name, class_name=class_.name
    )
    if induced_slot_definition.range != expected_target_class:
        raise MetadataModelTransformationError(
            f"Range of slot '{reference_slot_name}' in class"
            + f" '{class_.name}' does not match the reference specified in the"
            + " embedding profile"
        )

    # update the range if the target is another embedded class:
    if isinstance(target, EmbeddingProfile):
        slot_definition.range = target.target_class

    # set the target slot to inlined:
    slot_definition.inlined = True
    if induced_slot_definition.multivalued:
        slot_definition.inlined_as_list = True

    return slot_definition


def generated_embedded_class(
    *,
    schema_view: ExportableSchemaView,
    embedding_profile: EmbeddingProfile,
) -> ClassDefinition:
    """Generate an embedded class from an embedding profile."""

    class_to_embed = schema_view.get_class(embedding_profile.source_class)
    if not class_to_embed:
        raise MetadataModelTransformationError(
            f"Could not find class {embedding_profile.source_class} in model"
        )

    embedded_class = deepcopy(class_to_embed)

    # rename the class:
    embedded_class.name = embedding_profile.target_class

    if not embedded_class.slots:
        embedded_class.slots = []

    if not embedded_class.slot_usage:
        embedded_class.slot_usage = {}

    embedded_class.slot_usage = get_normalized_slot_usage(class_=embedded_class)

    for (
        reference_slot_name,
        target,
    ) in embedding_profile.embedded_references.items():
        if reference_slot_name not in embedded_class.slots:
            raise MetadataModelTransformationError(
                f"Slot '{reference_slot_name}' not found in class"
                + f" '{embedding_profile.source_class}'."
            )

        embedded_class.slot_usage[reference_slot_name] = get_embedded_reference_slot(
            class_=class_to_embed,
            schema_view=schema_view,
            reference_slot_name=reference_slot_name,
            target=target,
        )

    return embedded_class


def add_anchor_point_for_embedded_class(
    *,
    schema_view: ExportableSchemaView,
    embedding_profile: EmbeddingProfile,
) -> ExportableSchemaView:
    """Add an anchor point for an embedded class to the schema view."""

    identifier_slot = get_class_identifier(
        model=schema_view.export_model(), class_name=embedding_profile.source_class
    )

    if not identifier_slot:
        raise MetadataModelTransformationError(
            f"Could not find identifier slot for class {embedding_profile.source_class}"
        )
    anchor_point = AnchorPoint(
        root_slot=get_embedding_profile_root_slot(embedding_profile=embedding_profile),
        target_class=embedding_profile.target_class,
        identifier_slot=identifier_slot,
    )

    schema_view = add_anchor_point(
        schema_view=schema_view,
        anchor_point=anchor_point,
        description=embedding_profile.description,
    )

    return schema_view


def add_custom_embedded_class(
    *,
    model: MetadataModel,
    embedding_profile: EmbeddingProfile,
    anchor_points_by_target: dict[str, AnchorPoint],
    include_anchor_point: bool = True,
) -> MetadataModel:
    """Add a custom embedded class to a metadata model.
    If no anchor point is needed, specify `include_anchor_point=False`.
    """

    schema_view = model.schema_view
    embedded_class = generated_embedded_class(
        schema_view=schema_view,
        embedding_profile=embedding_profile,
    )
    schema_view.add_class(embedded_class)

    # add anchor point for embedded class:
    if include_anchor_point:
        schema_view = add_anchor_point_for_embedded_class(
            schema_view=schema_view, embedding_profile=embedding_profile
        )
    else:
        schema_view = disable_identifier_slot(
            schema_view=schema_view, class_name=embedded_class.name
        )

    model_modified = schema_view.export_model()

    # also prepare embedded classes for references:
    for target in embedding_profile.embedded_references.values():
        if isinstance(target, EmbeddingProfile):
            model_modified = add_custom_embedded_class(
                model=model_modified,
                embedding_profile=target,
                anchor_points_by_target=anchor_points_by_target,
                include_anchor_point=False,
            )

    return model_modified


def mark_anchored_classes_as_hidden(
    *, model: MetadataModel, anchor_points_by_target: dict[str, AnchorPoint]
) -> MetadataModel:
    """Mark all classes that are anchored as hidden. Returns a copy of the model."""

    schema_view = model.schema_view

    for anchor_point in anchor_points_by_target.values():
        schema_view = add_slot_usage_annotation(
            schema_view=schema_view,
            slot_name=anchor_point.root_slot,
            class_name=ROOT_CLASS,
            annotation_key="hidden",
            annotation_value=True,
        )

    return schema_view.export_model()


def add_custom_embedded_classes(
    *, model: MetadataModel, embedding_profiles: list[EmbeddingProfile]
) -> MetadataModel:
    """Add custom embedded classes to a metadata model."""

    anchor_points_by_target = get_anchors_points_by_target(model=model)

    # mark all existing anchored classes as hidden, so that only the custom embedded
    # classes are visible:
    model = mark_anchored_classes_as_hidden(
        model=model, anchor_points_by_target=anchor_points_by_target
    )

    for embedding_profile in embedding_profiles:
        model = add_custom_embedded_class(
            model=model,
            embedding_profile=embedding_profile,
            anchor_points_by_target=anchor_points_by_target,
        )

    return model



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Models to describe transformations and workflows."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable, Generic, Optional, TypeVar

from pydantic import BaseModel, Field, create_model, root_validator, validator

from metldata.custom_types import Json
from metldata.event_handling.models import SubmissionAnnotation

# shortcuts:
# pylint: disable=unused-import
from metldata.model_utils.assumptions import MetadataModelAssumptionError  # noqa: F401
from metldata.model_utils.essentials import MetadataModel


class MetadataModelTransformationError(RuntimeError):
    """Raised when a transformation failed when applied to the metadata model."""


class MetadataTransformationError(RuntimeError):
    """Raised when a transformation failed when applied to metadata."""


Config = TypeVar("Config", bound=BaseModel)


class MetadataTransformer(ABC, Generic[Config]):
    """A base class for a metadata transformer."""

    def __init__(
        self,
        *,
        config: Config,
        original_model: MetadataModel,
        transformed_model: MetadataModel,
    ):
        """Initialize the transformer with config params, the original model, and the
        transformed model."""

        self._config = config
        self._original_model = original_model
        self._transformed_model = transformed_model

    @abstractmethod
    def transform(self, *, metadata: Json, annotation: SubmissionAnnotation) -> Json:
        """Transforms metadata.

        Args:
            metadata: The metadata to be transformed.
            annotation: The annotation on the metadata.

        Raises:
            MetadataTransformationError:
                if the transformation fails.
        """
        ...


@dataclass(frozen=True)
class TransformationDefinition(Generic[Config]):
    """A model for describing a transformation."""

    config_cls: type[Config] = Field(
        ..., description="The config model of the transformation."
    )
    check_model_assumptions: Callable[[MetadataModel, Config], None] = Field(
        ...,
        description=(
            "A function that checks the assumptions made about the input model."
            "Raises a MetadataModelAssumptionError if the assumptions are not met."
        ),
    )
    transform_model: Callable[[MetadataModel, Config], MetadataModel] = Field(
        ...,
        description=(
            "A function to transform the model. Raises a"
            + " MetadataModelTransformationError if the transformation fails."
        ),
    )
    metadata_transformer_factory: type[MetadataTransformer[Config]] = Field(
        ...,
        description=(
            "A class for transforming metadata. Raises a MetadataTransformationError"
            "if the transformation fails."
        ),
    )


class WorkflowConfig(BaseModel, ABC):
    """A base class for workflow configs."""


class WorkflowStepBase(BaseModel, ABC):
    """A base class for workflow steps."""

    description: str = Field(..., description="A description of the step.")
    input: Optional[str] = Field(
        ...,
        description=(
            "The name of the workflow step from which the output is used as input"
            " for this step. If this is the first step, set to None."
        ),
    )

    class Config:
        """Config for the workflow step."""

        frozen = True


class WorkflowStep(WorkflowStepBase):
    """A single step in a transformation workflow."""

    transformation_definition: TransformationDefinition = Field(
        ...,
        description="The transformation to be executed in this step.",
    )


class WorkflowDefinition(BaseModel):
    """A definition of a transformation workflow."""

    description: str = Field(..., description="A description of the workflow.")
    steps: dict[str, WorkflowStep] = Field(
        ...,
        description=(
            "A dictionary of workflow steps. The keys are the names of the steps, and"
            + " the values are the workflow steps themselves."
        ),
    )
    artifacts: dict[str, str] = Field(
        ...,
        description=(
            "A dictionary of artifacts that are output by this workflow."
            + " The keys are the names of the artifacts, and the values are the names"
            + " of the workflow steps that output them."
        ),
    )

    # pylint: disable=no-self-argument
    @validator("steps", pre=False)
    def validate_step_references(
        cls, steps: dict[str, WorkflowStep]
    ) -> dict[str, WorkflowStep]:
        """Validate that workflow steps reference other existing steps as input.
        There should be exactly one step with input=None.
        """

        step_with_no_input_found = False

        for step_name, step in steps.items():
            if step.input is None:
                if step_with_no_input_found:
                    raise ValueError(
                        "There should be exactly one step with input=None. But multiple"
                        + " were found."
                    )
                step_with_no_input_found = True
                continue
            if step.input not in steps:
                raise ValueError(
                    f"Step {step.input} referenced in step {step_name} is not defined."
                )

        if not step_with_no_input_found:
            raise ValueError(
                "There should be exactly one step with input=None but none was found."
            )

        return steps

    @root_validator(pre=False)
    def validate_artifact_references(cls, values):
        """Validate that artifacts reference existing workflow steps."""

        steps = values.get("steps")
        if steps is None:
            raise ValueError("Steps are undefined.")
        artifacts = values.get("artifacts")
        if artifacts is None:
            raise ValueError("Artifacts are undefined.")

        for artifact_name, step_name in artifacts.items():
            if step_name not in steps:
                raise ValueError(
                    f"Step {step_name} referenced in artifact {artifact_name} is not defined."
                )

        return values

    @property
    def config_cls(self) -> type[WorkflowConfig]:
        """Get a config model containing the config requirements from all workflow
        steps."""

        step_configs = {
            step_name: (step.transformation_definition.config_cls, ...)
            for step_name, step in self.steps.items()
        }

        config_cls = create_model(  # type: ignore
            "SpecificWorkflowConfig",
            **step_configs,
            __base__=WorkflowConfig,
        )

        return config_cls

    @property
    def step_order(self) -> list[str]:
        """Get a list of step names in the order in which the steps should be executed."""

        step_order = list(self.steps)
        for _ in range(1000):
            # try to come up with a seqeuence of steps that satisfies all dependencies:
            for step_name in step_order:
                dependency = self.steps[step_name].input
                if dependency:
                    # check if the dependency is before the current step in the
                    # step order, if so, move the current step after the dependency:
                    self_position = step_order.index(step_name)
                    dependency_position = step_order.index(dependency)
                    if dependency_position > self_position:
                        step_order.remove(step_name)
                        step_order.insert(dependency_position + 1, step_name)
                        break
            else:
                # if we get here, we have found a valid step order and return it:
                return step_order

        raise RuntimeError(
            "Exceeded number of tries to resolve the step order."
            + " This is likely due to a circular dependency."
        )

    class Config:
        """Config for the workflow step."""

        frozen = True



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Models to describe slot merging instructions."""

from typing import Optional

from pydantic import BaseModel, Field, root_validator


class SlotMergeInstruction(BaseModel):
    """A model to describe slot merging instructions."""

    class_name: str = Field(..., description="The class to which the slots belong.")
    source_slots: list[str] = Field(
        ...,
        description="The slots that should be merged into the target slot.",
        min_items=2,
    )
    target_slot: str = Field(
        ..., description="The slot into which the source slots should be merged."
    )
    target_description: Optional[str] = Field(
        None,
        description="A description of the target slot.",
    )

    # pylint: disable=no-self-argument
    @root_validator()
    def validate_overlapping_slots(cls, values: dict) -> dict:
        """Validate that source and target slots do not overlap."""

        source_slots = set(values["source_slots"])
        target_slot = values["target_slot"]

        if target_slot in source_slots:
            raise ValueError("Source and target slots must not overlap.")

        return values



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Event Models."""


from pydantic import BaseModel, Field

from metldata.custom_types import SubmissionContent
from metldata.submission_registry.models import AccessionMap


class SubmissionAnnotation(BaseModel):
    """Annotation on a given metadata submission."""

    accession_map: AccessionMap = Field(
        ...,
        description=(
            "A map of user-specified id to system-generated accession for metadata"
            + " resources. Keys on the top level correspond to names of metadata classes."
            + " Keys on the second level correspond to user-specified aliases."
            + " Values on the second level correspond to system-generated accessions."
            + " Please note that the user-defined alias might only be unique within"
            + " the scope of the corresponding class and this submission. By contrast,"
            + " the system-generated accession is unique across all classes and"
            + " submissions."
        ),
    )


class SubmissionEventPayload(BaseModel):
    """Model for an event payload."""

    submission_id: str
    content: SubmissionContent
    annotation: SubmissionAnnotation



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""test the manipulate module"""

import pytest
from linkml_runtime.linkml_model.meta import SlotDefinition

from metldata.model_utils.identifiers import get_class_identifier
from metldata.model_utils.manipulate import (
    ClassNotFoundError,
    ClassSlotNotFoundError,
    add_slot_if_not_exists,
    add_slot_usage_annotation,
    delete_class_slot,
    disable_identifier_slot,
    upsert_class_slot,
)
from tests.fixtures.metadata_models import VALID_MINIMAL_METADATA_MODEL


def test_add_slot_if_not_exists_exists():
    """Test add_slot_if_not_exists if the slot exists."""

    # try to change existing files slot to string:
    new_slot = SlotDefinition(name="files", range="string")

    original_model = VALID_MINIMAL_METADATA_MODEL
    original_schema_view = original_model.schema_view

    updated_schema_view = add_slot_if_not_exists(
        schema_view=original_schema_view, new_slot=new_slot
    )
    updated_model = updated_schema_view.export_model()

    # nothing should have changed:
    assert updated_model == original_model


def test_add_slot_if_not_exists_not_exists():
    """Test add_slot_if_not_exists if the slot does not exist."""

    new_slot = SlotDefinition(name="test", range="string")

    original_model = VALID_MINIMAL_METADATA_MODEL
    original_schema_view = original_model.schema_view

    updated_schema_view = add_slot_if_not_exists(
        schema_view=original_schema_view, new_slot=new_slot
    )

    # make sure that the updated model differs from the original in the new slot:
    original_model_slots = set(original_schema_view.all_slots())
    updated_model_slots = set(updated_schema_view.all_slots())
    assert updated_model_slots.difference(original_model_slots) == {new_slot.name}


def test_upsert_class_slot_exists():
    """Test upsert_class_slot if the slot exists."""

    # change existing files slot to string in the Dataset class:
    class_name = "Dataset"
    new_slot = SlotDefinition(name="files", range="string")

    original_model = VALID_MINIMAL_METADATA_MODEL
    original_schema_view = original_model.schema_view

    updated_schema_view = upsert_class_slot(
        schema_view=original_schema_view, class_name=class_name, new_slot=new_slot
    )

    # induced_slot should have changed:
    updated_class_slot = updated_schema_view.induced_slot(
        slot_name=new_slot.name, class_name=class_name
    )
    assert updated_class_slot.range == "string"

    # check that the global slot definition has not changed:
    updated_global_slot = updated_schema_view.get_slot(slot_name=new_slot.name)
    assert updated_global_slot.range == "File"


def test_upsert_class_slot_not_exists():
    """Test upsert_class_slot if the slot does not exists."""

    class_name = "Dataset"
    new_slot = SlotDefinition(name="test", range="string")

    original_model = VALID_MINIMAL_METADATA_MODEL
    original_schema_view = original_model.schema_view

    updated_schema_view = upsert_class_slot(
        schema_view=original_schema_view, class_name=class_name, new_slot=new_slot
    )

    # check that the slot exists on the global and class level in an identical fashion:
    assert new_slot.name in updated_schema_view.class_slots(class_name=class_name)
    updated_class_slot = updated_schema_view.induced_slot(
        slot_name=new_slot.name, class_name=class_name
    )
    updated_global_slot = updated_schema_view.get_slot(slot_name=new_slot.name)
    assert updated_class_slot.range == updated_global_slot.range == new_slot.range


def test_delete_class_slot_happy():
    """Test the happy path of using delete_class_slot."""

    class_name = "Dataset"
    slot_name = "files"

    original_model = VALID_MINIMAL_METADATA_MODEL
    original_schema_view = original_model.schema_view

    updated_schema_view = delete_class_slot(
        schema_view=original_schema_view, class_name=class_name, slot_name=slot_name
    )

    # check that the slot does not exist on the class level anymore:
    assert slot_name not in updated_schema_view.class_slots(class_name=class_name)

    # check that the slot still exists on the global level:
    assert updated_schema_view.get_slot(slot_name=slot_name) is not None


def test_delete_class_slot_class_not_exists():
    """Test delete_class_slot if the class does not exist."""

    class_name = "DoesNotExist"
    slot_name = "files"

    original_model = VALID_MINIMAL_METADATA_MODEL
    original_schema_view = original_model.schema_view

    with pytest.raises(ClassNotFoundError):
        delete_class_slot(
            schema_view=original_schema_view, class_name=class_name, slot_name=slot_name
        )


def test_delete_class_slot_slot_not_exists():
    """Test delete_class_slot if the slot does not exist."""

    class_name = "Dataset"
    slot_name = "does_not_exist"

    original_model = VALID_MINIMAL_METADATA_MODEL
    original_schema_view = original_model.schema_view

    with pytest.raises(ClassSlotNotFoundError):
        delete_class_slot(
            schema_view=original_schema_view, class_name=class_name, slot_name=slot_name
        )


def test_add_slot_usage_annotation():
    """Test the happy path of using add_slot_usage_annotation."""

    class_name = "Dataset"
    slot_name = "files"
    annotation_key = "usage"
    annotation_value = "required"

    original_model = VALID_MINIMAL_METADATA_MODEL
    original_schema_view = original_model.schema_view

    updated_schema_view = add_slot_usage_annotation(
        schema_view=original_schema_view,
        class_name=class_name,
        slot_name=slot_name,
        annotation_key=annotation_key,
        annotation_value=annotation_value,
    )

    # check that annotations are correctly compiled to yaml level:
    model_dict = updated_schema_view.export_model().as_dict()
    assert (
        model_dict["classes"][class_name]["slot_usage"][slot_name]["annotations"][
            annotation_key
        ]["value"]
        == annotation_value
    )


def test_disable_identifier_slot():
    """Test the happy path of using disable_identifier_slot."""

    class_name = "Dataset"

    original_model = VALID_MINIMAL_METADATA_MODEL
    original_schema_view = original_model.schema_view

    updated_schema_view = disable_identifier_slot(
        schema_view=original_schema_view,
        class_name=class_name,
    )
    updated_model = updated_schema_view.export_model()

    # check that the slot is correctly disabled:
    assert get_class_identifier(model=updated_model, class_name=class_name) is None



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Models used to describe all inferred references based on existing references."""

from pydantic import BaseSettings, Field

from metldata.builtin_transformations.infer_references.reference import (
    InferredReference,
    ReferenceDetails,
)


class ReferenceInferenceConfig(BaseSettings):
    """Config containing inferred references for all classes of a metadata model in a
    dictionary-based representation and the option to translate that reference map into
    a list of InferredReferences."""

    inferred_ref_map: dict[str, dict[str, ReferenceDetails]] = Field(
        ...,
        description=(
            "A nested dictionary describing inferred references based"
            + " on existing references. On the first level keys refer to classes to"
            + " which inferred references should be added. On the second level, keys"
            + " refer to the names of the new slots of classes that hold the inferred"
            + " references. The values refer to the actual references details."
        ),
        example={
            "class_a": {
                "class_d": {
                    "path": "class_a(class_b)>class_b(class_d)>class_d",
                    "multivalued": False,
                },
                "class_c": {
                    "path": "class_a(class_b)>class_b<(class_c)class_c",
                    "multivalued": True,
                },
            },
            "class_b": {
                "class_c": {
                    "path": "class_b<(class_c)class_c",
                    "multivalued": True,
                }
            },
        },
    )

    @property
    def inferred_references(self) -> list[InferredReference]:
        """A list of inferred references."""

        inferred_refs: list[InferredReference] = []

        for source, slot_description in self.inferred_ref_map.items():
            for new_slot, reference_details in slot_description.items():
                target = reference_details.path.target
                inferred_refs.append(
                    InferredReference(
                        source=source,
                        target=target,
                        path=reference_details.path,
                        new_slot=new_slot,
                        multivalued=reference_details.multivalued,
                    )
                )

        return inferred_refs

    class Config:
        """Pydantic config."""

        extra = "forbid"



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""A tranformation for deleting slots from classes in a metadata model."""

from metldata.builtin_transformations.delete_slots.assumptions import (
    check_model_class_slots,
)
from metldata.builtin_transformations.delete_slots.config import SlotDeletionConfig
from metldata.builtin_transformations.delete_slots.metadata_transform import (
    delete_class_slots,
)
from metldata.builtin_transformations.delete_slots.model_transform import (
    delete_class_slots_from_model,
)
from metldata.event_handling.models import SubmissionAnnotation
from metldata.model_utils.anchors import get_anchors_points_by_target
from metldata.model_utils.assumptions import check_anchor_points
from metldata.model_utils.essentials import MetadataModel
from metldata.transform.base import Json, MetadataTransformer, TransformationDefinition


def check_model_assumptions(model: MetadataModel, config: SlotDeletionConfig):
    """Check that the classes and slots specified in the config exist in the model."""

    check_model_class_slots(model=model, class_slots=config.slots_to_delete)
    check_anchor_points(model=model, classes=list(config.slots_to_delete))


def transform_model(model: MetadataModel, config: SlotDeletionConfig) -> MetadataModel:
    """Delete slots from classes in the model."""

    return delete_class_slots_from_model(
        model=model, class_slots=config.slots_to_delete
    )


class SlotDeletionMetadataTransformer(MetadataTransformer[SlotDeletionConfig]):
    """Transformer for deleting slots from classes in a metadata model."""

    def __init__(
        self,
        config: SlotDeletionConfig,
        original_model: MetadataModel,
        transformed_model: MetadataModel,
    ):
        """Initialize the transformer."""

        super().__init__(
            config=config,
            original_model=original_model,
            transformed_model=transformed_model,
        )

        self._anchor_points_by_target = get_anchors_points_by_target(
            model=self._original_model
        )

    def transform(self, *, metadata: Json, annotation: SubmissionAnnotation) -> Json:
        """Transforms metadata.

        Args:
            metadata: The metadata to be transformed.
            annotation: The annotation on the metadata.

        Raises:
            MetadataTransformationError:
                if the transformation fails.
        """

        return delete_class_slots(
            metadata=metadata,
            slots_to_delete=self._config.slots_to_delete,
            anchor_points_by_target=self._anchor_points_by_target,
        )


SLOT_DELETION_TRANSFORMATION = TransformationDefinition[SlotDeletionConfig](
    config_cls=SlotDeletionConfig,
    check_model_assumptions=check_model_assumptions,
    transform_model=transform_model,
    metadata_transformer_factory=SlotDeletionMetadataTransformer,
)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for loading artifacts."""


from metldata.artifacts_rest.artifact_dao import ArtifactDaoCollection
from metldata.artifacts_rest.models import (
    ArtifactInfo,
    ArtifactResource,
    ArtifactResourceClass,
)
from metldata.custom_types import Json
from metldata.metadata_utils import get_resources_of_class, lookup_self_id


def extract_class_resources_from_artifact(
    *, artifact_content: Json, resource_class: ArtifactResourceClass
) -> list[ArtifactResource]:
    """Extract the resources from the given artifact content for the given class."""

    resource_jsons = get_resources_of_class(
        class_name=resource_class.name,
        global_metadata=artifact_content,
        anchor_points_by_target={resource_class.name: resource_class.anchor_point},
    )

    return [
        ArtifactResource(
            id_=lookup_self_id(
                resource=resource_json,
                identifier_slot=resource_class.anchor_point.identifier_slot,
            ),
            class_name=resource_class.name,
            content=resource_json,
        )
        for resource_json in resource_jsons
    ]


def extract_all_resources_from_artifact(
    artifact_content: Json, artifact_info: ArtifactInfo
) -> list[ArtifactResource]:
    """Extract all resources from the given artifact content for all resource classes
    as specified in the artifact info."""

    return [
        resource
        for resource_class in artifact_info.resource_classes.values()
        for resource in extract_class_resources_from_artifact(
            artifact_content=artifact_content,
            resource_class=resource_class,
        )
    ]


async def save_artifact_resource(
    *,
    resource: ArtifactResource,
    artifact_name: str,
    dao_collection: ArtifactDaoCollection
) -> None:
    """Save the given resource into the database using the given DAO collection."""

    dao = await dao_collection.get_dao(
        artifact_name=artifact_name, class_name=resource.class_name
    )
    await dao.upsert(resource)


async def load_artifact_resources(
    *,
    artifact_content: Json,
    artifact_info: ArtifactInfo,
    dao_collection: ArtifactDaoCollection
) -> None:
    """Load the resources from the given artifacts into the database using the given
    DAO collection.
    """

    resources = extract_all_resources_from_artifact(
        artifact_content=artifact_content, artifact_info=artifact_info
    )

    for resource in resources:
        await save_artifact_resource(
            resource=resource,
            artifact_name=artifact_info.name,
            dao_collection=dao_collection,
        )



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test the event_handling module."""

import json

import pytest
from hexkit.custom_types import Ascii, JsonObject
from hexkit.protocols.eventsub import EventSubscriberProtocol
from pydantic import BaseModel, Field

from metldata.event_handling.event_handling import FileSystemEventSubscriber
from tests.fixtures.event_handling import file_system_event_fixture  # noqa: F401
from tests.fixtures.event_handling import Event, FileSystemEventFixture

EXAMPLE_EVENTS = [
    Event(topic="topic1", type_="type1", key="key1", payload={"test1": "value1"}),
    # overwrites above event since same key:
    Event(topic="topic1", type_="type2", key="key1", payload={"test2": "value2"}),
    # other key thus will be present:
    Event(topic="topic1", type_="type1", key="key2", payload={"test3": "value3"}),
    # publish to other topic:
    Event(topic="topic2", type_="type1", key="key1", payload={"test4": "value4"}),
]


class ConsumedEvent(BaseModel):
    """Consumed event without the key."""

    topic: str
    type_: str
    payload: str = Field(..., description="JSON string of the event payload.")

    class Config:
        """Pydantic model configuration."""

        frozen = True


@pytest.mark.asyncio
async def test_pub_sub_workflow(
    file_system_event_fixture: FileSystemEventFixture,  # noqa: F811
):
    """Test a publish subscribe workflow using the FileSystemEventPublisher and
    FileSystemEventSubscriber."""

    # expected events to consume in topic1:
    expected_events = {
        ConsumedEvent(
            topic="topic1", type_="type2", payload=json.dumps({"test2": "value2"})
        ),
        ConsumedEvent(
            topic="topic1", type_="type1", payload=json.dumps({"test3": "value3"})
        ),
    }

    # publish events:
    await file_system_event_fixture.publish_events(EXAMPLE_EVENTS)

    # use event subscriber to consume events:
    class MockSubscriberTranslator(EventSubscriberProtocol):
        """A mock implementation of the EventSubscriberProtocol to track consumed
        events. Only consumes from topic1.

        Consumed events are captured in the consumed_events attribute.
        """

        def __init__(self):
            self.consumed_events: set[ConsumedEvent] = set()
            self.topics_of_interest = {"topic1"}
            self.types_of_interest = {"type1", "type2"}

        async def _consume_validated(
            self, *, payload: JsonObject, type_: Ascii, topic: Ascii
        ) -> None:
            self.consumed_events.add(
                ConsumedEvent(topic=topic, type_=type_, payload=json.dumps(payload))
            )

    translator = MockSubscriberTranslator()
    subscriber = FileSystemEventSubscriber(
        config=file_system_event_fixture.config, translator=translator
    )
    await subscriber.run()

    # check consumed events:
    assert translator.consumed_events == expected_events


@pytest.mark.asyncio
async def test_pub_collect_workflow(
    file_system_event_fixture: FileSystemEventFixture,  # noqa: F811
):
    """Test a publish collect workflow using the FileSystemEventPublisher and
    FileSystemEventCollector."""

    expected_events = EXAMPLE_EVENTS.copy()
    del expected_events[0]  # remove event with same key

    # publish events:
    await file_system_event_fixture.publish_events(EXAMPLE_EVENTS)

    # check published events with collector:
    file_system_event_fixture.expect_events(expected_events)
